# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0045: Multi-User Chat
# XEP-0249: Direct MUC Invitations

from __future__ import annotations

from typing import Any

import copy
import logging
import time
from collections import defaultdict
from collections.abc import Sequence
from datetime import timedelta

import nbxmpp
from gi.repository import GLib
from nbxmpp.const import Affiliation
from nbxmpp.const import InviteType
from nbxmpp.const import StatusCode
from nbxmpp.errors import StanzaError
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.modules.muc.util import MucInfoResult
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import Presence
from nbxmpp.structs import AffiliationResult
from nbxmpp.structs import CommonError
from nbxmpp.structs import CommonResult
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import MucConfigResult
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import VoiceRequest
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import events
from gajim.common import helpers
from gajim.common import types
from gajim.common.const import ClientState
from gajim.common.const import MUCJoinedState
from gajim.common.events import MucAdded
from gajim.common.events import MucDecline
from gajim.common.events import MucInvitation
from gajim.common.helpers import Observable
from gajim.common.helpers import to_user_string
from gajim.common.modules.base import BaseModule
from gajim.common.modules.bits_of_binary import store_bob_data
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.structs import MUCData
from gajim.common.structs import MUCPresenceData
from gajim.common.util.datetime import utc_now
from gajim.common.util.muc import format_private_group_name
from gajim.common.util.muc import get_default_muc_config
from gajim.common.util.muc import get_group_chat_nick

log = logging.getLogger("gajim.c.m.muc")


REJOIN_STATUS_CODES = {StatusCode.REMOVED_SERVICE_SHUTDOWN, StatusCode.REMOVED_ERROR}


class MUC(BaseModule):
    _nbxmpp_extends = "MUC"
    _nbxmpp_methods = [
        "get_affiliation",
        "set_role",
        "set_affiliation",
        "set_config",
        "set_subject",
        "cancel_config",
        "send_captcha",
        "cancel_captcha",
        "decline",
        "invite",
        "request_config",
        "request_voice",
        "approve_voice_request",
        "destroy",
        "request_disco_info",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._on_muc_user_presence,
                typ="available",
                ns=Namespace.MUC_USER,
                priority=49,
            ),
            StanzaHandler(
                name="presence",
                callback=self._on_muc_user_presence,
                typ="unavailable",
                ns=Namespace.MUC_USER,
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_muc_normal_user_message,
                typ="normal",
                ns=Namespace.MUC_USER,
                priority=49,
            ),
            StanzaHandler(
                name="presence",
                callback=self._on_error_presence,
                typ="error",
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_subject_change,
                typ="groupchat",
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_config_change,
                ns=Namespace.MUC_USER,
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_invite_or_decline,
                typ="normal",
                ns=Namespace.MUC_USER,
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_invite_or_decline,
                ns=Namespace.CONFERENCE,
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_captcha_challenge,
                ns=Namespace.CAPTCHA,
                priority=49,
            ),
            StanzaHandler(
                name="message",
                callback=self._on_voice_request,
                ns=Namespace.DATA,
                priority=49,
            ),
            StanzaHandler(
                name="message",
                typ="error",
                callback=self._message_error_received,
                priority=49,
            ),
        ]

        self._con.connect_signal("state-changed", self._on_client_state_changed)
        self._con.connect_signal("resume-failed", self._on_client_resume_failed)

        self._rejoin_muc: set[JID] = set()
        self._rejoin_timeouts: dict[JID, int] = {}
        self._muc_service_jid = None
        self._joined_users: defaultdict[JID, dict[str, MUCPresenceData]] = defaultdict(
            dict
        )
        self._mucs: dict[JID, MUCData] = {}
        self._muc_nicknames = {}
        self._muc_affiliations = AffiliationManager(self._log)
        self._muc_affiliations.multi_connect(
            {
                "affiliation-request-complete": self._on_affiliations_complete,
                "affiliations-changed": self._on_affiliations_changed,
            }
        )
        self._voice_requests: dict[GroupchatContact, list[VoiceRequest]] = defaultdict(
            list
        )
        self._sent_voice_requests: dict[JID, list[str]] = defaultdict(list)

    def _on_resume_failed(self, _client: types.Client, _signal_name: str) -> None:
        self._reset_presence()

    def _on_state_changed(
        self, _client: types.Client, _signal_name: str, state: ClientState
    ) -> None:
        if state.is_disconnected:
            self._reset_presence()

    @property
    def supported(self) -> bool:
        return self._muc_service_jid is not None

    @property
    def service_jid(self) -> JID | None:
        return self._muc_service_jid

    def pass_disco(self, info: DiscoInfo) -> None:
        if info.is_gateway:
            return

        for identity in info.identities:
            if identity.category != "conference":
                continue
            if identity.type != "text":
                continue
            if Namespace.MUC in info.features:
                self._log.info("Discovered MUC: %s", info.jid)
                self._muc_service_jid = info.jid
                raise nbxmpp.NodeProcessed

    def get_muc_data(self, room_jid: JID) -> MUCData | None:
        return self._mucs.get(room_jid)

    def set_password(self, room_jid: JID, password: str) -> None:
        muc_data = self.get_muc_data(room_jid)
        assert muc_data is not None
        muc_data.password = password

    def _get_mucs_with_state(self, states: list[MUCJoinedState]):
        return [muc for muc in self._mucs.values() if muc.state in states]

    def _set_muc_state(self, room_jid: JID, state: MUCJoinedState) -> None:
        try:
            muc = self._mucs[room_jid]
        except KeyError:
            raise ValueError("set_muc_state() called on unknown muc: %s" % room_jid)

        if muc.state == state:
            return

        self._log.info("Set MUC state: %s %s", room_jid, state)

        muc.state = state
        contact = self._get_contact(room_jid, groupchat=True)
        contact.notify("state-changed")

    def _reset_state(self) -> None:
        self._remove_all_timeouts()
        for muc in self._mucs.values():
            self._joined_users.pop(muc.jid, None)
            self._set_muc_state(muc.jid, MUCJoinedState.NOT_JOINED)
            room = self._get_contact(muc.jid)
            assert isinstance(room, GroupchatContact)
            room.set_not_joined()
            room.notify("room-left")

        self._joined_users.clear()

    def _create_muc_data(
        self,
        room_jid: JID,
        nick: str | None,
        password: str | None,
        config: dict[str, Any] | None,
    ) -> MUCData:
        if not nick:
            nick = get_group_chat_nick(self._account, room_jid)

        # Fetch data from bookmarks
        bookmark = self._con.get_module("Bookmarks").get_bookmark(room_jid)
        if bookmark is not None:
            if bookmark.password is not None:
                password = bookmark.password

        return MUCData(str(room_jid), nick, None, password, config)

    def join(
        self,
        jid: JID,
        nick: str | None = None,
        password: str | None = None,
        config: dict[str, Any] | None = None,
    ) -> None:
        if not app.account_is_available(self._account):
            return

        self._con.get_module("Contacts").add_contact(jid, groupchat=True)

        muc_data = self._mucs.get(jid)
        if muc_data is None:
            muc_data = self._create_muc_data(jid, nick, password, config)
            self._mucs[jid] = muc_data
            self._push_muc_added_event(jid)

        elif nick is not None:
            # Currently MUCData is never discarded so if it exists it contains
            # the nickname of a previous join. The user may chose now on a new
            # join a different nickname, so update MUCData here.
            muc_data.nick = nick

        if muc_data.state not in (
            MUCJoinedState.NOT_JOINED,
            MUCJoinedState.PASSWORD_REQUEST,
        ):
            self._log.warning("Can’t join MUC %s, state: %s", jid, muc_data.state)
            return

        disco_info = app.storage.cache.get_last_disco_info(muc_data.jid, max_age=60)
        if disco_info is None:
            self._set_muc_state(muc_data.jid, MUCJoinedState.BEFORE_JOINING)
            self._con.get_module("Discovery").disco_muc(
                muc_data.jid,
                callback=self._on_disco_result,  # type: ignore
            )
        else:
            self._join(muc_data)

    def create(self, jid: JID, config: dict[str, Any]) -> None:
        if not app.account_is_available(self._account):
            return

        self._con.get_module("Contacts").add_contact(jid, groupchat=True)
        muc_data = self._create_muc_data(jid, None, None, config)
        self._mucs[jid] = muc_data
        self._create(muc_data)
        self._push_muc_added_event(jid, select_chat=True)

    def _push_muc_added_event(self, jid: JID, select_chat: bool = False) -> None:

        app.ged.raise_event(
            MucAdded(account=self._account, jid=jid, select_chat=select_chat)
        )

    def _on_disco_result(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            assert error.jid is not None
            jid = error.jid
            self._log.info("Disco %s failed: %s", jid, error.get_text())
            muc_data = self._mucs.get(jid)
            if muc_data is None:
                self._log.info("Unable to find MUC data for %s", jid)
                return

            muc_data.error = "join-failed"
            error_text = helpers.to_user_string(error)
            muc_data.error_text = error_text
            self._set_muc_state(jid, MUCJoinedState.NOT_JOINED)
            room = self._get_contact(jid)
            room.notify("room-join-failed", error_text)
            return

        muc_data = self._mucs.get(result.info.jid)
        if muc_data is None:
            self._log.warning("MUC Data not found, join aborted")
            return
        self._join(muc_data)

    def _join(self, muc_data: MUCData) -> None:
        presence = self._con.get_module("Presence").get_presence(
            muc_data.occupant_jid,
            show=self._con.status,
            status=self._con.status_message,
        )

        muc_x = presence.setTag(Namespace.MUC + " x")
        muc_x.setTag("history", {"maxchars": "0"})

        if muc_data.password is not None:
            muc_x.setTagData("password", muc_data.password)

        self._log.info("Join MUC: %s", muc_data.jid)
        self._set_muc_state(muc_data.jid, MUCJoinedState.JOINING)
        self._con.send_stanza(presence)

    def _rejoin(self, room_jid: JID) -> bool:
        if not self._client.is_available():
            return True

        muc_data = self._mucs[room_jid]
        if muc_data.state.is_not_joined:
            self._log.info("Rejoin %s", room_jid)
            self._join(muc_data)
        return True

    def _create(self, muc_data: MUCData) -> None:
        presence = self._con.get_module("Presence").get_presence(
            muc_data.occupant_jid,
            show=self._con.status,
            status=self._con.status_message,
        )

        presence.setTag(Namespace.MUC + " x")

        self._log.info("Create MUC: %s", muc_data.jid)
        self._set_muc_state(muc_data.jid, MUCJoinedState.CREATING)
        self._con.send_stanza(presence)

    def leave(
        self,
        room_jid: JID,
        reason: str | None = None,
        unregister: bool = False,
    ) -> None:
        self._log.info("Leave MUC: %s", room_jid)

        self._con.get_module("Bookmarks").modify(room_jid, autojoin=False)

        muc_data = self._mucs.get(room_jid)
        if muc_data is None:
            return

        self._remove_rejoin_timeout(room_jid)

        if muc_data.state.is_not_joined:
            return

        self._con.get_module("Presence").send_presence(
            muc_data.occupant_jid, typ="unavailable", status=reason, caps=False
        )

        self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)
        room = self._get_contact(room_jid)
        assert isinstance(room, GroupchatContact)
        room.set_not_joined()
        room.notify("room-left")

        if not unregister:
            return

        contact = self._con.get_module("Contacts").get_contact(room_jid)
        assert isinstance(contact, GroupchatContact)
        self._con.get_module("Register").unregister(
            jid=contact.jid,
            callback=self._unregister_callback,
        )

    def _unregister_callback(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            self._log.error(
                "Error while unregistering from MUC: %s", to_user_string(error)
            )

    def abort_join(self, room_jid: JID) -> None:
        self._remove_rejoin_timeout(room_jid)

        muc = self._mucs[room_jid]
        self._con.get_module("Presence").send_presence(
            muc.occupant_jid, typ="unavailable", caps=False
        )

        self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)

    def configure_room(self, room_jid: JID) -> None:
        self._nbxmpp("MUC").request_config(room_jid, callback=self._on_room_config)

    def _on_room_config(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.info(error)
            assert error.jid is not None
            room = self._get_contact(error.jid.new_as_bare())
            error_text = helpers.to_user_string(error)
            room.notify("room-config-failed", error_text)
            return

        assert isinstance(result, MucConfigResult)
        self._log.info("Configure room: %s", result.jid)

        muc_data = self._mucs[result.jid]
        assert result.form is not None
        self._apply_config(result.form, muc_data.config)
        self.set_config(result.jid, result.form, callback=self._on_config_result)

    @staticmethod
    def _apply_config(
        form: SimpleDataForm, config: dict[str, Any] | None = None
    ) -> None:
        default_config = get_default_muc_config()
        if config is not None:
            default_config.update(config)
        for var, value in default_config.items():
            try:
                field = form[var]
            except KeyError:
                pass
            else:
                field.value = value  # type: ignore

    def _on_config_result(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.info(error)
            assert error.jid is not None
            room = self._get_contact(error.jid.new_as_bare())
            error_text = helpers.to_user_string(error)
            room.notify("room-config-failed", error_text)
            return

        assert isinstance(result, CommonResult)
        self._con.get_module("Discovery").disco_muc(
            result.jid,
            callback=self._on_disco_result_after_config,  # type: ignore
        )

    def _on_disco_result_after_config(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.info("Disco %s failed: %s", error.jid, error.get_text())
            return

        assert isinstance(result, MucInfoResult)
        jid = result.info.jid
        assert jid is not None
        muc_data = self._mucs[jid]
        self._room_join_complete(muc_data)

        self._log.info("Configuration finished: %s", jid)

        room = self._get_contact(jid.new_as_bare())
        event = events.MUCRoomConfigFinished(timestamp=utc_now())

        assert isinstance(room, GroupchatContact)
        app.storage.events.store(room, event)
        room.notify("room-config-finished", event)

    def _request_affiliation_list(self, room_jid: JID) -> None:
        # Don't request affiliation outcast, it is rarley needed and
        # can be requested on demand rather than on each join

        self._muc_affiliations.start_affiliation_request(room_jid)
        for affiliation in ("owner", "admin", "member"):
            self.get_affiliation(
                room_jid,
                affiliation,
                callback=self._on_affiliations_received,
                user_data=(room_jid, affiliation),
            )

    def _on_affiliations_received(self, task: Task) -> None:
        jid, affiliation = task.get_user_data()
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.info("Affiliation request failed: %s %s", jid, error)
            self._muc_affiliations.process_affiliation_result(jid, affiliation, None)
            return

        assert isinstance(result, AffiliationResult)
        self._muc_affiliations.process_affiliation_result(jid, affiliation, result)

    def _on_affiliations_complete(
        self, manager: AffiliationManager, _signal: str, jid: JID
    ) -> None:
        self._update_muc_fallback_name(jid)
        room = self._get_contact(jid)
        assert isinstance(room, GroupchatContact)
        room.notify("room-affiliations-complete")

    def _on_affiliations_changed(
        self,
        manager: AffiliationManager,
        _signal: str,
        jid: JID,
        *args: Any,
    ) -> None:
        self._update_muc_fallback_name(jid)

    def get_affiliations(
        self, room_jid: JID, *, include_outcast: bool = False
    ) -> dict[str, list[JID]]:

        return self._muc_affiliations.get_users(
            room_jid, include_outcast=include_outcast
        )

    def update_presence(self) -> None:
        # Only update joined MUCs, ejabberd seems to have a bug
        # that when we send a presence after the initial join presence
        # it will overwrite our join presence with it.
        # This can lead to receiving MUC history because only the join
        # presence has <history maxchars="0" />
        mucs = self._get_mucs_with_state([MUCJoinedState.JOINED])

        status, message, idle = self._con.get_presence_state()
        for muc_data in mucs:
            self._con.get_module("Presence").send_presence(
                muc_data.occupant_jid, show=status, status=message, idle_time=idle
            )

    def change_nick(self, room_jid: JID, new_nick: str) -> None:
        status, message, _idle = self._con.get_presence_state()
        self._con.get_module("Presence").send_presence(
            f"{room_jid}/{new_nick}", show=status, status=message
        )

    def _on_error_presence(
        self,
        _con: types.NBXMPPClient,
        _stanza: Presence,
        properties: PresenceProperties,
    ) -> None:

        assert properties.jid is not None
        assert properties.error is not None
        room_jid = properties.jid.new_as_bare()
        muc_data = self._mucs.get(room_jid)
        if muc_data is None:
            return

        room = self._get_contact(room_jid)

        if muc_data.state == MUCJoinedState.JOINING:
            if properties.error.condition == "conflict":
                self._remove_rejoin_timeout(room_jid)
                muc_data.nick += "_"
                self._log.info(
                    "Nickname conflict: %s change to %s", muc_data.jid, muc_data.nick
                )
                self._join(muc_data)

            elif properties.error.condition == "not-authorized":
                self._remove_rejoin_timeout(room_jid)
                self._set_muc_state(room_jid, MUCJoinedState.PASSWORD_REQUEST)
                room.notify("room-password-required")

            else:
                self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)
                if room_jid not in self._rejoin_muc:
                    muc_data.error = "join-failed"
                    assert isinstance(properties.error, CommonError)
                    muc_data.error_text = helpers.to_user_string(properties.error)
                    room.notify("room-join-failed", muc_data.error_text)

        elif muc_data.state == MUCJoinedState.CREATING:
            self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)
            muc_data.error = "creation-failed"
            assert isinstance(properties.error, CommonError)
            muc_data.error_text = helpers.to_user_string(properties.error)
            room.notify("room-creation-failed", muc_data.error_text)

        elif muc_data.state == MUCJoinedState.CAPTCHA_REQUEST:
            self._set_muc_state(room_jid, MUCJoinedState.CAPTCHA_FAILED)
            self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)
            muc_data.error = "captcha-failed"
            assert isinstance(properties.error, CommonError)
            muc_data.error_text = helpers.to_user_string(properties.error)
            room.notify("room-captcha-error", muc_data.error_text)

        elif muc_data.state == MUCJoinedState.CAPTCHA_FAILED:
            self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)

        else:
            error = helpers.to_user_string(properties.error)
            event = events.MUCRoomPresenceError(timestamp=utc_now(), error=error)
            assert isinstance(room, GroupchatContact)
            app.storage.events.store(room, event)
            room.notify("room-presence-error", event)

    def _on_muc_user_presence(
        self, _con: types.NBXMPPClient, stanza: Presence, properties: PresenceProperties
    ) -> None:

        assert properties.type is not None
        assert properties.muc_jid is not None
        room_jid = properties.muc_jid
        if room_jid not in self._mucs:
            self._log.warning("Presence from unknown MUC")
            self._log.warning(stanza)
            return

        muc_data = self._mucs[room_jid]
        assert properties.jid is not None
        occupant = self._get_contact(properties.jid, groupchat=True)
        assert isinstance(occupant, GroupchatParticipant)
        room = self._get_contact(properties.jid.new_as_bare())
        assert isinstance(room, GroupchatContact)

        timestamp = utc_now()

        if properties.muc_destroyed is not None:
            self._log.info("MUC destroyed: %s", room_jid)
            self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)
            self._con.get_module("Bookmarks").remove(room_jid)
            room.set_not_joined()

            event = events.MUCRoomDestroyed(
                timestamp=timestamp,
                reason=properties.muc_destroyed.reason,
                alternate=properties.muc_destroyed.alternate,
            )
            assert isinstance(room, GroupchatContact)
            app.storage.events.store(room, event)

            room.notify("room-destroyed", event)
            return

        if properties.is_nickname_changed:
            assert properties.muc_user is not None
            muc_user_nick = properties.muc_user.nick
            assert muc_user_nick is not None

            if properties.is_muc_self_presence:
                muc_data.nick = muc_user_nick
                self._con.get_module("Bookmarks").modify(
                    muc_data.jid, nick=muc_data.nick
                )

            initiator = "Server" if properties.is_nickname_modified else "User"
            self._log.info(
                "%s nickname changed: %s to %s",
                initiator,
                properties.jid,
                muc_user_nick,
            )

            # We receive only the unavailable presence here, so we take
            # the current presence and create a new contact with it, before we
            # update the presence.
            new_occupant = room.add_resource(muc_user_nick)
            new_occupant.set_presence(occupant.presence)
            self._joined_users[room.jid][muc_user_nick] = occupant.presence

            presence = self._process_user_presence(properties)
            self._process_occupant_presence_change(properties, presence, occupant)

            event = events.MUCNicknameChanged(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                new_name=new_occupant.name,
                old_name=occupant.name,
            )

            assert isinstance(room, GroupchatContact)
            app.storage.events.store(room, event)

            room.notify("user-nickname-changed", event, occupant, new_occupant)
            return

        is_joined = self._is_user_joined(properties.jid)
        if not is_joined and properties.type.is_available:
            if properties.is_muc_self_presence:
                self._log.info("Self presence: %s", properties.jid)
                if muc_data.state == MUCJoinedState.JOINING:
                    if room.supports(Namespace.OCCUPANT_ID):
                        muc_data.occupant_id = properties.occupant_id

                    if (
                        properties.is_nickname_modified
                        or muc_data.nick != properties.muc_nickname
                    ):
                        assert properties.muc_nickname is not None
                        muc_data.nick = properties.muc_nickname
                        self._log.info(
                            "Server modified nickname to: %s", properties.muc_nickname
                        )

                elif muc_data.state == MUCJoinedState.CREATING:
                    if properties.is_new_room:
                        muc_data.occupant_id = properties.occupant_id
                        self.configure_room(room_jid)

            presence = self._process_user_presence(properties)
            self._process_occupant_presence_change(properties, presence, occupant)
            return

        if properties.is_muc_self_presence and properties.is_kicked:
            self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)
            room.set_not_joined()

            assert properties.muc_user is not None

            event = events.MUCRoomKicked(
                timestamp=timestamp,
                status_codes=properties.muc_status_codes,
                reason=properties.muc_user.reason,
                actor=properties.muc_user.actor,
            )
            assert isinstance(room, GroupchatContact)
            app.storage.events.store(room, event)
            room.notify("room-kicked", event)

            status_codes = properties.muc_status_codes or set()
            if REJOIN_STATUS_CODES & status_codes:
                self._start_rejoin_timeout(room_jid)
            return

        if properties.is_muc_self_presence and properties.type.is_unavailable:
            # Its not a kick, so this is the reflection of our own
            # unavailable presence, because we left the MUC
            return

        if properties.jid.resource is None:
            # prosody allows broadcasting "unavailable" presences from
            # offline room members
            return

        try:
            presence = self._process_user_presence(properties)
        except KeyError:
            # Sometimes it seems to happen that we get unavailable presence
            # from occupants we don’t know
            log.warning("Unexpected presence received")
            log.warning(stanza)
            return

        self._process_occupant_presence_change(properties, presence, occupant)

    def _process_occupant_presence_change(
        self,
        properties: PresenceProperties,
        presence: MUCPresenceData,
        occupant: GroupchatParticipant,
    ) -> None:

        assert properties.muc_jid is not None
        timestamp = utc_now()

        if not occupant.is_available and presence.available:
            event = events.MUCUserJoined(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                nick=occupant.name,
                status_codes=properties.muc_status_codes,
            )

            if occupant.room.is_joined or properties.is_muc_self_presence:
                # Don’t store initial presences on join
                app.storage.events.store(occupant.room, event)

            occupant.update_presence(presence)
            occupant.notify("user-joined", event)
            return

        if not presence.available:
            assert properties.muc_user is not None

            event = events.MUCUserLeft(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                nick=occupant.name,
                status_codes=properties.muc_status_codes,
                reason=properties.muc_user.reason,
                actor=properties.muc_user.actor,
            )

            app.storage.events.store(occupant.room, event)

            status_codes = properties.muc_status_codes or set()
            if (
                StatusCode.REMOVED_AFFILIATION_CHANGE in status_codes
                and presence.real_jid is not None
            ):
                self._muc_affiliations.change_user_affiliation(
                    properties.muc_jid, presence.affiliation, presence.real_jid
                )
                event = events.MUCUserAffiliationChanged(
                    timestamp=timestamp,
                    is_self=properties.is_muc_self_presence,
                    nick=occupant.name,
                    affiliation=presence.affiliation,
                    reason=properties.muc_user.reason,
                    actor=properties.muc_user.actor,
                )
                occupant.notify("user-affiliation-changed", event)

            occupant.update_presence(presence)
            occupant.notify("user-left", event)
            return

        signals_and_events: list[tuple[str, Any]] = []

        if occupant.affiliation != presence.affiliation:
            assert properties.muc_user is not None

            event = events.MUCUserAffiliationChanged(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                nick=occupant.name,
                affiliation=presence.affiliation,
                reason=properties.muc_user.reason,
                actor=properties.muc_user.actor,
            )

            app.storage.events.store(occupant.room, event)
            if presence.real_jid is not None:
                self._muc_affiliations.change_user_affiliation(
                    properties.muc_jid, presence.affiliation, presence.real_jid
                )
            signals_and_events.append(("user-affiliation-changed", event))

        if occupant.role != presence.role:
            assert properties.muc_user is not None

            event = events.MUCUserRoleChanged(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                nick=occupant.name,
                role=presence.role,
                reason=properties.muc_user.reason,
                actor=properties.muc_user.actor,
            )

            app.storage.events.store(occupant.room, event)
            signals_and_events.append(("user-role-changed", event))

        if occupant.status != presence.status or occupant.show != presence.show:
            assert properties.show is not None
            event = events.MUCUserStatusShowChanged(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                nick=occupant.name,
                status=properties.status,
                show_value=properties.show.value,
            )

            app.storage.events.store(occupant, event)
            app.storage.events.store(occupant.room, event)
            signals_and_events.append(("user-status-show-changed", event))

        if occupant.hats != presence.hats:
            if presence.hats:
                hats = [h.title for h in presence.hats.get_hats()]
            else:
                hats = None
            event = events.MUCUserHatsChanged(
                timestamp=timestamp,
                is_self=properties.is_muc_self_presence,
                nick=occupant.name,
                hats=hats,
            )

            app.storage.events.store(occupant, event)
            app.storage.events.store(occupant.room, event)
            signals_and_events.append(("user-hats-changed", event))

        occupant.update_presence(presence)
        for signal, event in signals_and_events:
            occupant.notify(signal, event)

    def _on_muc_normal_user_message(
        self, _con: types.NBXMPPClient, stanza: Presence, properties: MessageProperties
    ) -> None:
        if not properties.muc_user or not properties.muc_user.jid:
            return

        if not properties.muc_user.affiliation and not properties.muc_user.role:
            return

        room_jid = properties.muc_jid
        if room_jid not in self._mucs:
            self._log.warning("Message from unknown MUC")
            self._log.warning(stanza)
            return

        assert room_jid is not None
        room = self._get_contact(room_jid)
        assert isinstance(room, GroupchatContact)

        if properties.muc_user.nick is not None:
            nick = properties.muc_user.nick
        else:
            contact = self._get_contact(properties.muc_user.jid)
            if isinstance(contact, BareContact):
                nick = contact.name
            else:
                nick = properties.muc_user.jid.to_user_string()

        assert properties.muc_user.affiliation is not None
        event = events.MUCAffiliationChanged(
            timestamp=utc_now(),
            nick=nick,
            affiliation=properties.muc_user.affiliation,
            jid=properties.muc_user.jid,
        )
        app.storage.events.store(room, event)
        self._muc_affiliations.change_user_affiliation(
            room_jid, properties.muc_user.affiliation, properties.muc_user.jid
        )
        room.notify("room-affiliation-changed", event)

    def _process_user_presence(self, properties: PresenceProperties) -> MUCPresenceData:
        assert properties.jid is not None
        jid = properties.jid
        group_contact = self._client.get_module("Contacts").get_contact(
            jid.bare, groupchat=True
        )
        occupant_id = None
        if group_contact.supports(Namespace.OCCUPANT_ID):
            occupant_id = properties.occupant_id

        muc_presence = MUCPresenceData.from_presence(properties, occupant_id)
        assert jid.resource is not None
        if not muc_presence.available:
            self._joined_users[jid.new_as_bare()].pop(jid.resource)
        else:
            self._joined_users[jid.new_as_bare()][jid.resource] = muc_presence
        return muc_presence

    def _is_user_joined(self, jid: JID) -> bool:
        assert jid.resource is not None
        try:
            self._joined_users[jid.new_as_bare()][jid.resource]
        except KeyError:
            return False
        return True

    def get_joined_users(self, jid: JID) -> list[str]:
        return list(self._joined_users[jid].keys())

    def _start_rejoin_timeout(self, room_jid: JID) -> None:
        self._remove_rejoin_timeout(room_jid)
        self._rejoin_muc.add(room_jid)
        self._log.info("Start rejoin timeout for: %s", room_jid)
        id_ = GLib.timeout_add_seconds(10, self._rejoin, room_jid)
        self._rejoin_timeouts[room_jid] = id_

    def _remove_rejoin_timeout(self, room_jid: JID) -> None:
        self._rejoin_muc.discard(room_jid)
        id_ = self._rejoin_timeouts.get(room_jid)
        if id_ is not None:
            self._log.info("Remove rejoin timeout for: %s", room_jid)
            GLib.source_remove(id_)
            del self._rejoin_timeouts[room_jid]

    def _on_subject_change(
        self, _con: types.NBXMPPClient, _stanza: Message, properties: MessageProperties
    ) -> None:

        if not properties.is_muc_subject:
            return

        assert properties.jid is not None
        room_jid = properties.jid.new_as_bare()

        muc_data = self._mucs.get(room_jid)
        if muc_data is None:
            self._log.warning("No MUCData found for %s", room_jid)
            return

        room = self._get_contact(JID.from_string(room_jid))
        assert isinstance(room, GroupchatContact)

        old_subject = muc_data.subject

        muc_subject = properties.muc_subject
        assert muc_subject is not None
        muc_data.subject = muc_subject

        if muc_subject.timestamp is None:
            muc_subject = muc_subject._replace(timestamp=time.time())

        if old_subject is None:
            muc_data.last_subject_timestamp = time.time()
            room.notify("room-subject", muc_subject)

        elif old_subject.text != muc_subject.text:
            # Check if we already showed that subject (rejoin)
            muc_data.last_subject_timestamp = time.time()
            room.notify("room-subject", muc_subject)

        if muc_data.state == MUCJoinedState.JOINING:
            self._room_join_complete(muc_data)
            self._request_affiliation_list(muc_data.jid)
            room.notify("room-joined")

        raise nbxmpp.NodeProcessed

    def cancel_password_request(self, room_jid: JID) -> None:
        self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)

    def _room_join_complete(self, muc_data: MUCData) -> None:
        # Reset errors from previous tries, otherwise when we are
        # disconnected from the room, the ChatFunctionPage will be shown
        muc_data.error = None
        muc_data.error_text = None

        self._set_muc_state(muc_data.jid, MUCJoinedState.JOINED)
        self._remove_rejoin_timeout(muc_data.jid)

        # We successfully joined a MUC, set add bookmark with autojoin
        self._con.get_module("Bookmarks").add_or_modify(
            muc_data.jid, autojoin=True, password=muc_data.password, nick=muc_data.nick
        )

        disco_info = app.storage.cache.get_last_disco_info(muc_data.jid)
        assert disco_info is not None
        if disco_info.has_mam_2:
            self._con.get_module("MAM").request_archive_on_muc_join(muc_data.jid)

    def _update_muc_fallback_name(self, jid: JID) -> None:
        contact = self._get_contact(jid, groupchat=True)
        assert isinstance(contact, GroupchatContact)
        if contact.muc_context != "private":
            app.storage.archive.set_contact_value(
                self._account, jid, "fallback_name", None
            )
            contact.notify("name-update")
            return

        affiliations = self.get_affiliations(jid)
        members: set[JID] = set()
        members.update(*affiliations.values())
        members.discard(self._get_own_bare_jid())

        occupants = app.storage.archive.get_occupant_by_jids(
            self._account, jid, list(members), max_age=timedelta(minutes=60)
        )

        nicks: list[str] = []
        for real_jid, occupant in occupants.items():
            if nick := occupant.nickname:
                nicks.append(nick)
                members.discard(real_jid)

        for member in members:
            nicks.append(member.localpart or member.domain)

        muc_name = format_private_group_name(nicks)
        if not muc_name:
            self._log.warning("Unable to calculate private muc name for %s", jid)
            return

        app.storage.archive.set_contact_value(
            self._account, jid, "fallback_name", muc_name
        )
        self._log.info("Update fallback name for %s to %s", jid, muc_name)
        contact.notify("name-update")

    def request_voice(self, jid: JID) -> None:
        message_id = self._nbxmpp("MUC").request_voice(jid)
        self._sent_voice_requests[jid].append(message_id)

    def _on_voice_request(
        self, _con: types.NBXMPPClient, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_voice_request:
            return

        assert properties.jid is not None
        room = self._get_contact(properties.jid.new_as_bare())
        assert isinstance(room, GroupchatContact)
        assert properties.voice_request is not None

        self._voice_requests[room].append(properties.voice_request)
        room.notify("room-voice-request", properties)

        raise nbxmpp.NodeProcessed

    def get_voice_requests(self, contact: GroupchatContact) -> list[VoiceRequest]:

        return self._voice_requests.get(contact, [])

    def approve_voice_request(
        self, contact: GroupchatContact, voice_request: VoiceRequest
    ) -> None:

        self._voice_requests[contact].remove(voice_request)
        self._nbxmpp("MUC").approve_voice_request(contact.jid, voice_request)

    def decline_voice_request(
        self, contact: GroupchatContact, voice_request: VoiceRequest
    ) -> None:

        self._voice_requests[contact].remove(voice_request)

    def _message_error_received(
        self,
        _con: types.NBXMPPClient,
        stanza: nbxmpp.Message,
        properties: MessageProperties,
    ) -> None:

        remote_jid = properties.remote_jid
        assert remote_jid is not None
        message_id = properties.id
        if message_id is None:
            self._log.warning("Received error without message id")
            self._log.warning(stanza)
            return

        # We handle only voice request errors for now
        requests = self._sent_voice_requests.get(remote_jid, [])
        if message_id not in requests:
            return

        room = self._get_contact(remote_jid, groupchat=True)
        assert isinstance(room, GroupchatContact)

        assert properties.error is not None
        error = helpers.to_user_string(properties.error)
        event = events.MUCRoomVoiceRequestError(
            timestamp=utc_now(), message_id=message_id, error=error
        )

        app.storage.events.store(room, event)

        room.notify("room-voice-request-error", event)
        raise nbxmpp.NodeProcessed

    def _on_captcha_challenge(
        self, _con: types.NBXMPPClient, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_captcha_challenge:
            return

        if properties.is_mam_message:
            # Some servers store captcha challenges in MAM, don’t process them
            self._log.warning("Ignore captcha challenge received from MAM")
            raise nbxmpp.NodeProcessed

        assert properties.jid is not None
        muc_data = self._mucs.get(properties.jid)
        if muc_data is None:
            return

        if muc_data.state != MUCJoinedState.JOINING:
            self._log.warning(
                "Received captcha request but state != %s", MUCJoinedState.JOINING
            )
            return

        self._log.info("Captcha challenge received from %s", properties.jid)

        assert properties.captcha is not None
        store_bob_data(properties.captcha.bob_data)
        muc_data.captcha_id = properties.id
        muc_data.captcha_form = properties.captcha.form

        self._set_muc_state(properties.jid, MUCJoinedState.CAPTCHA_REQUEST)
        self._remove_rejoin_timeout(properties.jid)

        room = self._get_contact(properties.jid.new_as_bare())
        room.notify("room-captcha-challenge")

        raise nbxmpp.NodeProcessed

    def cancel_captcha(self, room_jid: JID) -> None:
        muc_data = self._mucs.get(room_jid)
        if muc_data is None:
            return

        if muc_data.captcha_id is None:
            self._log.warning("No captcha message id available")
            return
        self._nbxmpp("MUC").cancel_captcha(room_jid, muc_data.captcha_id)
        self._set_muc_state(room_jid, MUCJoinedState.CAPTCHA_FAILED)
        self._set_muc_state(room_jid, MUCJoinedState.NOT_JOINED)

    def send_captcha(self, room_jid: JID, form_node: SimpleDataForm) -> None:
        self._set_muc_state(room_jid, MUCJoinedState.JOINING)
        self._nbxmpp("MUC").send_captcha(
            room_jid, form_node, callback=self._on_captcha_result
        )

    def _on_captcha_result(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            assert error.jid is not None
            jid = error.jid
            muc_data = self._mucs.get(jid)
            if muc_data is None:
                return
            self._set_muc_state(jid, MUCJoinedState.CAPTCHA_FAILED)
            room = self._get_contact(jid)
            error_text = helpers.to_user_string(error)
            room.notify("room-captcha-error", error_text)

    def _on_config_change(
        self, _con: types.NBXMPPClient, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_muc_config_change:
            return

        self._log.info(
            "Received config change: %s %s",
            properties.muc_jid,
            properties.muc_status_codes,
        )

        assert properties.muc_status_codes is not None
        event = events.MUCRoomConfigChanged(
            timestamp=utc_now(), status_codes=properties.muc_status_codes
        )

        assert properties.muc_jid is not None
        room = self._get_contact(properties.muc_jid)

        assert isinstance(room, GroupchatContact)
        app.storage.events.store(room, event)

        room.notify("room-config-changed", event)

        raise nbxmpp.NodeProcessed

    def _on_invite_or_decline(
        self, _con: types.NBXMPPClient, _stanza: Message, properties: MessageProperties
    ) -> None:

        assert properties.from_ is not None
        if properties.from_.bare_match(self._get_own_bare_jid()):
            # Invite/Decline sent by us and received via MAM or Carbons
            # TODO: It could make sense to delete a pending invite
            # if another device of ours decline the invite
            self._log.warning("Ignored MUC Invite/Decline sent by other device")
            return

        if properties.muc_decline is not None:
            data = properties.muc_decline
            contact = self._get_contact(data.muc, groupchat=True)

            self._log.info(
                "Invite declined from: %s, reason: %s", data.from_, data.reason
            )

            app.ged.raise_event(MucDecline(account=self._account, **data._asdict()))
            raise nbxmpp.NodeProcessed

        if properties.muc_invite is not None:
            data = properties.muc_invite
            contact = self._get_contact(data.muc, groupchat=True)
            assert isinstance(contact, GroupchatContact)

            self._log.info("Invite from: %s, to: %s", data.from_, data.muc)

            if contact.is_joined:
                # We are already in groupchat. Ignore invitation
                self._log.info("We are already in this room")
                raise nbxmpp.NodeProcessed

            self._con.get_module("Discovery").disco_muc(
                data.muc,
                request_vcard=True,
                callback=self._on_disco_result_after_invite,  # type: ignore
                user_data=data,  # type: ignore
            )

            raise nbxmpp.NodeProcessed

    def _on_disco_result_after_invite(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.warning(error)
            return

        invite_data = task.get_user_data()
        app.ged.raise_event(
            MucInvitation(
                account=self._account, info=result.info, **invite_data._asdict()
            )
        )

    def invite(
        self, room: JID, jid: JID, reason: str | None = None, continue_: bool = False
    ) -> str:

        room_contact = self._get_contact(room, groupchat=True)
        assert isinstance(room_contact, GroupchatContact)
        disco = room_contact.get_disco()
        assert disco is not None

        if disco.muc_is_members_only:
            self_contact = room_contact.get_self()
            assert self_contact is not None
            affiliation = self_contact.affiliation
            admin = affiliation.is_owner or affiliation.is_admin
            if admin:
                user_affiliation = self._muc_affiliations.get_user_affiliation(
                    room, jid
                )
                if user_affiliation in ("outcast", None):
                    self.set_affiliation(room, {jid: {"affiliation": "member"}})
                type_ = InviteType.DIRECT
            else:
                type_ = InviteType.MEDIATED
        else:
            type_ = InviteType.DIRECT

        password = self._mucs[room].password
        self._log.info("Invite %s to %s", jid, room)
        return self._nbxmpp("MUC").invite(room, jid, password, reason, continue_, type_)

    def moderate_messages(
        self, namespace: str, jid: JID, moderate_ids: Sequence[str], reason: str | None
    ) -> None:

        for moderation_id in moderate_ids:
            self._nbxmpp("MUC").moderate_message(namespace, jid, moderation_id, reason)

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: ClientState
    ) -> None:
        if state.is_disconnected:
            self._reset_state()

    def _on_client_resume_failed(
        self, _client: types.Client, _signal_name: str
    ) -> None:
        self._reset_state()

    def _remove_all_timeouts(self) -> None:
        for room_jid in list(self._rejoin_timeouts.keys()):
            self._remove_rejoin_timeout(room_jid)

    def cleanup(self) -> None:
        BaseModule.cleanup(self)
        self._remove_all_timeouts()


class AffiliationManager(Observable):
    def __init__(self, log: logging.LoggerAdapter[Any]) -> None:
        Observable.__init__(self)
        self._log = log
        self._affiliations: dict[JID, dict[str, list[JID]]] = defaultdict(
            lambda: defaultdict(list)
        )
        self._running_affiliation_requests: dict[JID, set[str]] = {}

    def start_affiliation_request(self, muc_jid: JID) -> None:
        self._running_affiliation_requests[muc_jid] = {"member", "admin", "owner"}

    def process_affiliation_result(
        self, muc_jid: JID, affiliation: str, result: AffiliationResult | None
    ) -> None:

        if muc_jid not in self._running_affiliation_requests:
            self._log.info("Unexpected affiliation request for %s", muc_jid)
            return

        if result is None:
            self._log.info(
                "AffiliationManager: Remove affiliation: %s %s", muc_jid, affiliation
            )
            try:
                del self._affiliations[muc_jid][affiliation]
            except KeyError:
                pass

        else:
            user_jids = [jid.new_as_bare() for jid in result.users]
            self._log.info(
                "AffiliationManager: Add result: %s %s %s",
                result.jid,
                affiliation,
                list(map(str, user_jids)),
            )
            self._affiliations[result.jid][affiliation] = user_jids

        open_requests = self._running_affiliation_requests[muc_jid]
        open_requests.discard(affiliation)
        if not open_requests:
            self._log.info("AffiliationManager: Request complete for %s", muc_jid)
            self._running_affiliation_requests.pop(muc_jid)
            self.notify("affiliation-request-complete", muc_jid)

    def _add_user_affiliation(
        self, muc_jid: JID, affiliation: Affiliation, user_real_jid: JID
    ) -> None:
        self._log.info(
            "AffiliationManager: Add user: %s %s %s",
            muc_jid,
            affiliation,
            user_real_jid,
        )
        self._affiliations[muc_jid][affiliation.value].append(user_real_jid)

    def _remove_user_affiliation(self, muc_jid: JID, user_real_jid: JID) -> None:
        self._log.info("AffiliationManager: Remove user: %s %s", muc_jid, user_real_jid)
        for jids in self._affiliations.get(muc_jid, {}).values():
            if user_real_jid in jids:
                jids.remove(user_real_jid)
                return

    def change_user_affiliation(
        self, muc_jid: JID, affiliation: Affiliation, user_real_jid: JID
    ) -> None:
        self._log.info(
            "AffiliationManager: Change user affiliation: %s %s %s",
            muc_jid,
            affiliation,
            user_real_jid,
        )

        self._remove_user_affiliation(muc_jid, user_real_jid)
        if affiliation != Affiliation.NONE:
            self._add_user_affiliation(muc_jid, affiliation, user_real_jid)
        self.notify("affiliations-changed", muc_jid, user_real_jid, affiliation)

    def get_users(
        self, jid: JID, *, include_outcast: bool = False
    ) -> dict[str, list[JID]]:

        affiliations = copy.deepcopy(self._affiliations[jid])
        if not include_outcast:
            affiliations.pop("outcast", None)
        return affiliations

    def get_user_affiliation(self, muc_jid: JID, jid: JID) -> str | None:
        affiliations = self._affiliations[muc_jid]
        for affiliation in ("member", "admin", "owner", "outcast"):
            if jid in affiliations.get(affiliation, []):
                return affiliation
        return None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import NamedTuple

from gajim.common.iana.time_zones import WINDOWS_ZONES
from gajim.common.iana.time_zones import ZONES


class ZoneData(NamedTuple):
    key: str
    name: str

    @classmethod
    def from_key(cls, key: str) -> ZoneData:
        name = key.replace("_", " ")
        return cls(key, name)


def get_zone_data(key: str) -> ZoneData | None:
    if key not in ZONES:
        key = WINDOWS_ZONES.get(key, "")

    if not key:
        return None

    return ZoneData.from_key(key)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import colorsys
import functools
import hashlib
import logging
import math
from collections import defaultdict
from math import pi
from pathlib import Path

import cairo
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Pango
from gi.repository import PangoCairo
from nbxmpp.protocol import JID

import gajim.common.storage.archive.models as mod
from gajim.common import app
from gajim.common import configpaths
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.const import StyleAttr
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatOfflineParticipant
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.contacts import ResourceContact
from gajim.common.util.classes import Singleton
from gajim.common.util.image import get_pixbuf_from_file
from gajim.common.util.image import scale_with_ratio
from gajim.common.util.muc import get_groupchat_name
from gajim.common.util.status import get_client_status
from gajim.common.util.status import get_global_show

from gajim.gtk.const import DEFAULT_WORKSPACE_COLOR
from gajim.gtk.util.icons import load_icon_surface
from gajim.gtk.util.misc import convert_surface_to_texture
from gajim.gtk.util.misc import convert_texture_to_surface
from gajim.gtk.util.misc import get_first_graphemes
from gajim.gtk.util.styling import convert_rgb_string_to_float
from gajim.gtk.util.styling import get_contact_color
from gajim.gtk.util.styling import get_css_show_class
from gajim.gtk.util.styling import make_rgba
from gajim.gtk.util.styling import rgba_to_float
from gajim.gtk.util.styling import text_to_color

log = logging.getLogger("gajim.gtk.avatar")


AvatarCacheT = dict[
    JID | str, dict[tuple[int, int, str | None, str | None], Gdk.Texture]
]
OccupantAvatarCacheT = dict[JID, dict[tuple[str, int, int], Gdk.Texture]]

CIRCLE_RATIO = 0.18
CIRCLE_FILL_RATIO = 0.80


def generate_avatar_letter(text: str) -> str:
    return get_first_graphemes(text.lstrip(), 1).upper()


def generate_avatar(
    letters: str,
    color: tuple[float, float, float],
    size: int,
    scale: int,
    check_contrast: bool = False,
) -> cairo.ImageSurface:
    # Get color for nickname with XEP-0392
    color_r, color_g, color_b = color

    # Set up colors and size
    size = size * scale

    width = size
    height = size
    font_size = size * 0.4

    # Set up surface
    surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
    context = cairo.Context(surface)

    context.set_source_rgb(color_r, color_g, color_b)
    context.rectangle(0, 0, width, height)
    context.fill()

    # Draw letters
    layout = PangoCairo.create_layout(context)
    layout.set_text(generate_avatar_letter(letters))

    description = Pango.FontDescription()
    description.set_family("Sans")
    description.set_size(int(font_size * Pango.SCALE))

    layout.set_font_description(description)

    _ink_rect, logical_rect = layout.get_extents()
    layout_width = logical_rect.width / Pango.SCALE
    layout_height = logical_rect.height / Pango.SCALE
    x_pos = (width - layout_width) / 2
    y_pos = (height - layout_height) / 2

    context.move_to(x_pos, y_pos)

    text_color = (0.95, 0.95, 0.95)
    if check_contrast:
        _hue, luminance, _saturation = colorsys.rgb_to_hls(*color)
        if luminance > 0.6:
            # Use dark text color if background luminance threshold is surpassed
            text_color = (0.2, 0.2, 0.2)
    context.set_source_rgb(*text_color)
    context.set_operator(cairo.Operator.OVER)
    PangoCairo.show_layout(context, layout)

    return context.get_target()


@functools.cache
def generate_default_avatar(
    letter: str,
    color: tuple[float, float, float],
    size: int,
    scale: int,
    style: str = "circle",
) -> cairo.ImageSurface:
    surface = generate_avatar(letter, color, size, scale)
    surface = clip(surface, style)
    surface.set_device_scale(scale, scale)
    return surface


@functools.cache
def make_workspace_avatar(
    letter: str,
    color: tuple[float, float, float],
    size: int,
    scale: int,
    style: str = "round-corners",
) -> Gdk.Texture:
    surface = generate_avatar(letter, color, size, scale, check_contrast=True)
    surface.set_device_scale(scale, scale)
    return convert_surface_to_texture(clip(surface, style))


def add_transport_to_avatar(
    surface: cairo.ImageSurface,
    transport_icon: str,
) -> cairo.ImageSurface:
    scale = surface.get_device_scale()[0]
    width = surface.get_width()
    height = surface.get_height()

    transport_surface = load_icon_surface(
        transport_icon, math.floor(width / scale * CIRCLE_RATIO * 2), int(scale)
    )
    if transport_surface is None:
        return surface

    new_surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
    new_surface.set_device_scale(*surface.get_device_scale())

    context = cairo.Context(new_surface)
    context.set_source_surface(surface, 0, 0)
    context.paint()

    width = width / scale

    clip_radius = width * CIRCLE_RATIO
    center_x = clip_radius
    center_y = clip_radius

    context.set_source_rgb(255, 255, 255)
    context.set_operator(cairo.Operator.CLEAR)
    context.arc(center_x, center_y, clip_radius, 0, 2 * pi)
    context.fill()

    clip_radius *= CIRCLE_FILL_RATIO
    t_width = transport_surface.get_width() / scale
    t_height = transport_surface.get_height() / scale

    context.set_source_surface(
        transport_surface, center_x - t_width / 2, center_y - t_height / 2
    )
    context.set_operator(cairo.Operator.OVER)
    context.arc(center_x, center_y, clip_radius, 0, 2 * pi)
    context.fill()

    return context.get_target()


def add_status_to_avatar(surface: cairo.ImageSurface, show: str) -> cairo.ImageSurface:
    width = surface.get_width()
    height = surface.get_height()

    new_surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
    new_surface.set_device_scale(*surface.get_device_scale())

    scale = surface.get_device_scale()[0]

    context = cairo.Context(new_surface)
    context.set_source_surface(surface, 0, 0)
    context.paint()

    # Correct height and width for scale
    width = width / scale
    height = height / scale

    clip_radius = width * CIRCLE_RATIO
    center_x = width - clip_radius
    center_y = height - clip_radius

    context.set_source_rgb(255, 255, 255)
    context.set_operator(cairo.Operator.CLEAR)
    context.arc(center_x, center_y, clip_radius, 0, 2 * pi)
    context.fill()

    css_color = get_css_show_class(show)
    color = convert_rgb_string_to_float(
        app.css_config.get_value(css_color, StyleAttr.COLOR)
    )

    show_radius = clip_radius * CIRCLE_FILL_RATIO

    context.set_source_rgb(*color)
    context.set_operator(cairo.Operator.OVER)
    context.arc(center_x, center_y, show_radius, 0, 2 * pi)
    context.fill()

    if show == "dnd":
        line_length = clip_radius / 2
        context.move_to(center_x - line_length, center_y)
        context.line_to(center_x + line_length, center_y)

        context.set_source_rgb(255, 255, 255)
        context.set_line_width(clip_radius / 4)
        context.stroke()

    return context.get_target()


def merge_avatars(avatar1: Gdk.Texture, avatar2: Gdk.Texture) -> Gdk.Texture:
    """
    Includes avatar2 in the bottom right corner of avatar1
    """
    new_surface = cairo.ImageSurface(
        cairo.Format.ARGB32,
        avatar1.get_width(),
        avatar1.get_height(),
    )

    context = cairo.Context(new_surface)
    context.set_source_surface(convert_texture_to_surface(avatar1), 0, 0)
    context.paint()

    context.scale(0.5, 0.5)
    context.set_source_rgb(255, 255, 255)
    context.set_operator(cairo.Operator.CLEAR)
    context.arc(
        avatar1.get_width() * 1.5,
        avatar1.get_height() * 1.5,
        avatar1.get_width() / CIRCLE_FILL_RATIO / 2,
        0,
        2 * pi,
    )
    context.fill()

    context = cairo.Context(new_surface)
    context.scale(0.5, 0.5)
    context.set_source_surface(
        convert_texture_to_surface(avatar2),
        avatar1.get_width(),
        avatar1.get_height(),
    )
    context.paint()
    return convert_surface_to_texture(new_surface)


@functools.lru_cache(maxsize=128)
def get_show_circle(
    show: str | types.PresenceShowT, size: int, scale: int
) -> cairo.ImageSurface:
    if not isinstance(show, str):
        show = show.value

    width = size * scale
    height = width

    surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
    surface.set_device_scale(scale, scale)
    context = cairo.Context(surface)

    css_color = get_css_show_class(show)
    color = convert_rgb_string_to_float(
        app.css_config.get_value(css_color, StyleAttr.COLOR)
    )

    center = size / 2
    radius = size / 3

    context.set_source_rgb(*color)
    context.set_operator(cairo.Operator.OVER)
    context.arc(center, center, radius, 0, 2 * pi)
    context.fill()

    if show == "dnd":
        line_length = radius * 0.65
        context.move_to(center - line_length, center)
        context.line_to(center + line_length, center)

        context.set_source_rgb(255, 255, 255)
        context.set_line_width(size / 10)
        context.stroke()

    return context.get_target()


def fit(surface: cairo.ImageSurface, size: int) -> cairo.ImageSurface:
    width = surface.get_width()
    height = surface.get_height()
    if width == height:
        return surface

    # Fit any non-square image by:
    # 1. cutting a square from the original surface
    # 2. scaling the square to the desired size
    min_size = min(width, height)
    factor = size / min_size

    square_surface = square(surface, min_size)

    new_surface = cairo.ImageSurface(cairo.Format.ARGB32, size, size)
    new_surface.set_device_scale(*surface.get_device_scale())
    context = cairo.Context(new_surface)
    context.scale(factor, factor)
    context.set_source_surface(square_surface, 0, 0)
    context.paint()
    return context.get_target()


def square(surface: cairo.ImageSurface, size: int) -> cairo.ImageSurface:
    width = surface.get_width()
    height = surface.get_height()
    if width == height:
        return surface

    new_surface = cairo.ImageSurface(cairo.Format.ARGB32, size, size)
    new_surface.set_device_scale(*surface.get_device_scale())
    context = cairo.Context(new_surface)

    scale = surface.get_device_scale()[0]

    if width == size:
        x_pos = 0
        y_pos = (size - height) / 2 / scale
    else:
        y_pos = 0
        x_pos = (size - width) / 2 / scale

    context.set_source_surface(surface, x_pos, y_pos)
    context.paint()
    return context.get_target()


def clip(surface: cairo.ImageSurface, mode: str) -> cairo.ImageSurface:
    if mode == "circle":
        return clip_circle(surface)
    if mode == "round-corners":
        return round_corners(surface)
    raise ValueError("clip mode unknown: %s" % mode)


def clip_circle(surface: cairo.ImageSurface) -> cairo.ImageSurface:
    new_surface = cairo.ImageSurface(
        cairo.Format.ARGB32, surface.get_width(), surface.get_height()
    )

    new_surface.set_device_scale(*surface.get_device_scale())
    context = cairo.Context(new_surface)
    context.set_source_surface(surface, 0, 0)

    width = surface.get_width()
    height = surface.get_height()
    scale = surface.get_device_scale()[0]
    radius = width / 2 / scale

    context.arc(width / 2 / scale, height / 2 / scale, radius, 0, 2 * pi)

    context.clip()
    context.paint()

    return context.get_target()


def round_corners(surface: cairo.ImageSurface) -> cairo.ImageSurface:
    width = surface.get_width()
    height = surface.get_height()
    scale = surface.get_device_scale()[0]

    new_surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
    new_surface.set_device_scale(*surface.get_device_scale())
    context = cairo.Context(new_surface)
    context.set_source_surface(surface, 0, 0)

    width = width / scale
    height = height / scale

    radius = 9
    degrees = pi / 180

    context.new_sub_path()
    context.arc(width - radius, radius, radius, -90 * degrees, 0 * degrees)
    context.arc(width - radius, height - radius, radius, 0 * degrees, 90 * degrees)  # noqa: E501
    context.arc(radius, height - radius, radius, 90 * degrees, 180 * degrees)
    context.arc(radius, radius, radius, 180 * degrees, 270 * degrees)
    context.close_path()
    context.clip()

    context.paint()

    return context.get_target()


def convert_to_greyscale(surface: cairo.ImageSurface) -> cairo.ImageSurface:
    context = cairo.Context(surface)
    context.set_operator(cairo.Operator.HSL_COLOR)
    context.set_source_rgb(1, 1, 1)
    context.rectangle(0, 0, surface.get_width(), surface.get_height())
    context.fill()
    context.set_operator(cairo.Operator.ATOP)
    context.set_source_rgba(1, 1, 1, 0.5)
    context.rectangle(0, 0, surface.get_width(), surface.get_height())
    context.fill()
    return context.get_target()


class AvatarStorage(metaclass=Singleton):
    def __init__(self):
        self._cache: AvatarCacheT = defaultdict(dict)
        self._occupant_cache: OccupantAvatarCacheT = defaultdict(dict)

    def invalidate_cache(self, jid: JID | str) -> None:
        self._cache.pop(jid, None)

    def remove_avatar(self, contact: types.ChatContactT) -> None:
        if not contact.avatar_sha:
            return
        path = self.get_avatar_path(contact.avatar_sha)
        if path is None:
            return
        path.unlink(missing_ok=True)
        self._cache.pop(contact.jid, None)

    def get_texture(
        self,
        contact: (
            BareContact
            | GroupchatContact
            | GroupchatParticipant
            | GroupchatOfflineParticipant
        ),
        size: int,
        scale: int,
        show: str | None = None,
        default: bool = False,
        transport_icon: str | None = None,
        style: str = "circle",
    ) -> Gdk.Texture:
        jid = contact.jid

        if not default:
            texture = self._cache[jid].get((size, scale, show, transport_icon))
            if texture is not None:
                return texture

            surface = self._get_avatar_from_storage(
                contact.avatar_sha, size, scale, style
            )
            if surface is not None:
                if show is not None:
                    surface = add_status_to_avatar(surface, show)

                if transport_icon is not None:
                    surface = add_transport_to_avatar(surface, transport_icon)

                texture = convert_surface_to_texture(surface)
                self._cache[jid][(size, scale, show, transport_icon)] = texture
                return texture

        name = contact.name
        color = get_contact_color(contact)
        letter = generate_avatar_letter(name)
        surface = generate_default_avatar(letter, color, size, scale, style=style)
        if show is not None:
            surface = add_status_to_avatar(surface, show)

        if transport_icon is not None:
            surface = add_transport_to_avatar(surface, transport_icon)

        texture = convert_surface_to_texture(surface)
        self._cache[jid][(size, scale, show, transport_icon)] = texture
        return texture

    def get_muc_texture(
        self,
        account: str,
        jid: JID,
        size: int,
        scale: int,
        default: bool = False,
        transport_icon: str | None = None,
        style: str = "circle",
    ) -> Gdk.Texture:
        if not default:
            texture = self._cache[jid].get((size, scale, None, transport_icon))
            if texture is not None:
                return texture

            avatar_sha = app.storage.archive.get_contact_value(
                account, jid, "avatar_sha"
            )
            if avatar_sha is not None:
                surface = self.surface_from_filename(avatar_sha, size, scale)
                if surface is not None:
                    # clip first to avoid clipping the transport icon
                    surface = clip(surface, style)

                    if transport_icon is not None:
                        surface = add_transport_to_avatar(surface, transport_icon)

                    texture = convert_surface_to_texture(surface)
                    self._cache[jid][(size, scale, None, transport_icon)] = texture
                    return texture

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact_if_exists(jid)
        assert contact is not None
        assert not isinstance(contact, ResourceContact)

        name = get_groupchat_name(client, jid)
        color = get_contact_color(contact)
        letter = generate_avatar_letter(name)
        surface = generate_default_avatar(letter, color, size, scale, style)
        if transport_icon is not None:
            surface = add_transport_to_avatar(surface, transport_icon)

        texture = convert_surface_to_texture(surface)
        self._cache[jid][(size, scale, None, transport_icon)] = texture
        return texture

    def get_occupant_texture(
        self,
        jid: JID,
        occupant: mod.Occupant | str,
        size: int,
        scale: int,
        style: str = "circle",
    ) -> Gdk.Texture:
        if isinstance(occupant, str):
            key = occupant
            nickname = occupant
            avatar_sha = None
            real_remote = None
        else:
            key = occupant.id
            # There are edge cases with MUC PMs where we don't
            # have a nickname stored for an occupant
            nickname = occupant.nickname or "Unknown"
            avatar_sha = occupant.avatar_sha
            real_remote = occupant.real_remote

        texture = self._occupant_cache[jid].get((key, size, scale))
        if texture is not None:
            return texture

        surface = self._get_avatar_from_storage(avatar_sha, size, scale, style)
        if surface is not None:
            texture = convert_surface_to_texture(surface)
            self._occupant_cache[jid][(key, size, scale)] = texture
            return texture

        if real_remote is not None:
            color = text_to_color(str(real_remote.jid))
        else:
            color = text_to_color(nickname)

        letter = generate_avatar_letter(nickname)
        surface = generate_default_avatar(letter, color, size, scale, style=style)

        texture = convert_surface_to_texture(surface)
        self._occupant_cache[jid][(key, size, scale)] = texture
        return texture

    def get_own_avatar_texture(
        self, account: str, size: int, scale: int, add_show: bool = False
    ) -> Gdk.Texture | None:
        client = app.get_client(account)
        self_contact = client.get_module("Contacts").get_contact(
            client.get_own_jid().bare
        )
        assert isinstance(self_contact, BareContact)
        return self_contact.get_avatar(size, scale, add_show=add_show)

    def get_workspace_texture(
        self, workspace_id: str, size: int, scale: int
    ) -> Gdk.Texture | None:
        texture = self._cache[workspace_id].get((size, scale, None, None))
        if texture is not None:
            return texture

        name = app.settings.get_workspace_setting(workspace_id, "name")
        color = app.settings.get_workspace_setting(workspace_id, "color")
        avatar_sha = app.settings.get_workspace_setting(workspace_id, "avatar_sha")
        if avatar_sha:
            surface = self.surface_from_filename(avatar_sha, size, scale)
            if surface is not None:
                surface = clip(surface, "round-corners")
                texture = convert_surface_to_texture(surface)
                return texture

        rgba = make_rgba(color or DEFAULT_WORKSPACE_COLOR)
        texture = make_workspace_avatar(name, rgba_to_float(rgba), size, scale)

        self._cache[workspace_id][(size, scale, None, None)] = texture
        return texture

    @staticmethod
    def get_account_button_texture(
        account: str | None, size: int, scale: int, connectivity_issues: bool
    ) -> Gdk.Texture:
        width = size * scale
        height = width

        if connectivity_issues:
            surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
            context = cairo.Context(surface)

            context.set_source_rgb(0.75, 0.75, 0.75)
            context.rectangle(0, 0, width, height)
            context.fill()
            icon_surface = load_icon_surface("lucide-cloud-off", int(size * 0.7), scale)
            if icon_surface is not None:
                pos = (width - width * 0.7) / 2
                context.set_source_surface(icon_surface, pos, pos)
                context.paint_with_alpha(0.6)

            surface.set_device_scale(scale, scale)
            surface = clip_circle(context.get_target())
            return convert_surface_to_texture(surface)

        if account is not None:
            jid = app.get_jid_from_account(account)
            client = app.get_client(account)
            contact = client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            return app.app.avatar_storage.get_texture(
                contact, size, scale, get_client_status(account)
            )

        # Paint default avatar on grey background (incl. show)
        surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
        context = cairo.Context(surface)

        context.set_source_rgb(0.75, 0.75, 0.75)
        context.rectangle(0, 0, width, height)
        context.fill()

        icon_surface = load_icon_surface("lucide-user", int(size * 0.7), scale)
        if icon_surface is not None:
            pos = (width - width * 0.7) / 2
            context.set_source_surface(icon_surface, pos, pos)
            context.paint_with_alpha(0.6)

        surface.set_device_scale(scale, scale)
        surface = clip_circle(context.get_target())
        surface = add_status_to_avatar(surface, get_global_show())
        return convert_surface_to_texture(surface)

    @staticmethod
    def get_gajim_circle_icon(size: int, scale: int) -> Gdk.Texture:
        # Paint activity icon on grey background
        width = size * scale
        height = width

        surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
        context = cairo.Context(surface)

        context.set_source_rgb(1, 1, 1)
        context.rectangle(0, 0, width, height)
        context.fill()

        icon_surface = load_icon_surface("gajim", int(size * 0.6), scale)
        if icon_surface is not None:
            pos = (width - width * 0.6) / 2
            context.set_source_surface(icon_surface, pos, pos)
            context.paint()

        surface.set_device_scale(scale, scale)
        surface = clip_circle(context.get_target())
        return convert_surface_to_texture(surface)

    @staticmethod
    def _load_for_publish(path: str) -> tuple[bool, bytes] | None:
        pixbuf = get_pixbuf_from_file(path)
        if pixbuf is None:
            return None

        width = pixbuf.get_width()
        height = pixbuf.get_height()
        if width > AvatarSize.PUBLISH or height > AvatarSize.PUBLISH:
            # Scale only down, never up
            width, height = scale_with_ratio(AvatarSize.PUBLISH, width, height)
            pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)

        assert pixbuf is not None
        return pixbuf.save_to_bufferv("png", [], [])

    @staticmethod
    def save_avatar(data: bytes) -> str | None:
        """
        Save an avatar to the harddisk

        :param data:  bytes

        returns SHA1 value of the avatar or None on error
        """

        sha = hashlib.sha1(data).hexdigest()
        path = configpaths.get("AVATAR") / sha
        try:
            with open(path, "wb") as output_file:
                output_file.write(data)
        except Exception:
            log.exception("Storing avatar failed")
            return None
        return sha

    @staticmethod
    def get_avatar_path(filename: str) -> Path | None:
        path = configpaths.get("AVATAR") / filename
        if not path.is_file():
            return None
        return path

    def avatar_exists(self, filename: str) -> bool:
        return self.get_avatar_path(filename) is not None

    def surface_from_filename(
        self, filename: str, size: int, scale: int
    ) -> cairo.ImageSurface | None:
        size = size * scale
        path = self.get_avatar_path(filename)
        if path is None:
            return None

        pixbuf = get_pixbuf_from_file(path, size)
        if pixbuf is None:
            return None

        if pixbuf.get_n_channels() == 3:
            cairo_format = cairo.Format.RGB24
        else:
            cairo_format = cairo.Format.ARGB32

        surface = cairo.ImageSurface(
            cairo_format, pixbuf.get_width(), pixbuf.get_height()
        )

        context = cairo.Context(surface)

        Gdk.cairo_set_source_pixbuf(context, pixbuf, 0, 0)
        context.paint()

        return fit(context.get_target(), size)

    def _get_avatar_from_storage(
        self,
        sha: str | None,
        size: int,
        scale: int,
        style: str,
    ) -> cairo.ImageSurface | None:
        if sha is None:
            return None

        surface = self.surface_from_filename(sha, size, scale)
        if surface is None:
            return None
        return clip(surface, style)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import logging

from gi.repository import Adw
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.protocol import JID
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import ged
from gajim.common.const import MUC_DISCO_ERRORS
from gajim.common.events import AccountConnected
from gajim.common.events import AccountDisconnected
from gajim.common.ged import EventHelper
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _
from gajim.common.util.jid import validate_jid
from gajim.common.util.muc import get_random_muc_localpart

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.groupchat_creation")


@Gtk.Template(string=get_ui_string("groupchat_creation.ui"))
class CreateGroupchatWindow(GajimAppWindow, EventHelper):
    __gtype_name__ = "CreateGroupchatWindow"

    _stack: Gtk.Stack = Gtk.Template.Child()

    _account_group: Adw.PreferencesGroup = Gtk.Template.Child()
    _account_dropdown: GajimDropDown[str] = Gtk.Template.Child()

    _settings_listbox: Gtk.ListBox = Gtk.Template.Child()
    _name_entry_row: Adw.EntryRow = Gtk.Template.Child()
    _description_entry_row: Adw.EntryRow = Gtk.Template.Child()
    _advanced_expander_row: Adw.ExpanderRow = Gtk.Template.Child()
    _public_radio: Gtk.CheckButton = Gtk.Template.Child()
    _private_radio: Gtk.CheckButton = Gtk.Template.Child()
    _address_entry_row: Adw.EntryRow = Gtk.Template.Child()

    _error_label: Gtk.Label = Gtk.Template.Child()
    _info_label: Gtk.Label = Gtk.Template.Child()

    _spinner: Adw.Spinner = Gtk.Template.Child()
    _create_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self, account: str | None) -> None:
        GajimAppWindow.__init__(
            self,
            name="CreateGroupchat",
            title=_("Create Group Chat"),
            default_width=500,
            default_height=550,
        )
        EventHelper.__init__(self)

        self._connect(
            self._account_dropdown,
            "notify::selected",
            self._on_account_changed,
        )

        self._connect(
            self._address_entry_row, "changed", self._on_address_entry_changed
        )
        self._connect(self._create_button, "clicked", self._on_create_clicked)

        self._account = account
        self._destroyed: bool = False

        self.register_events(
            [
                ("account-connected", ged.GUI2, self._on_account_state),
                ("account-disconnected", ged.GUI2, self._on_account_state),
            ]
        )

        if app.get_number_of_connected_accounts() == 0:
            # This can happen under rare circumstances
            self._stack.set_visible_child_name("no-connection")
            self._create_button.set_sensitive(False)
            return

        self._update_accounts(account)
        self._create_button.grab_focus()

    def _cleanup(self) -> None:
        # https://gitlab.gnome.org/GNOME/gtk/-/issues/7992
        self._name_entry_row.props.enable_emoji_completion = False
        self._description_entry_row.props.enable_emoji_completion = False

        self._account_dropdown.run_destroy()
        self.unregister_events()
        self._destroyed = True

    def _on_account_state(self, _event: AccountConnected | AccountDisconnected) -> None:
        any_account_connected = app.get_number_of_connected_accounts() > 0
        if any_account_connected:
            self._stack.set_visible_child_name("create")
            self._create_button.set_sensitive(True)
        else:
            self._stack.set_visible_child_name("no-connection")
            self._create_button.set_sensitive(False)

        selected_account = self._account_dropdown.get_selected_key()
        self._update_accounts(selected_account)

    def _update_accounts(self, account: str | None = None) -> None:
        accounts = app.get_enabled_accounts_with_labels(connected_only=True)

        account_data = dict(accounts)
        self._account_dropdown.set_data(account_data)

        self._account_group.set_visible(len(accounts) != 1)

        if account is not None and self._account_dropdown.has_key(account):
            self._account_dropdown.select_key(account)
        else:
            self._account_dropdown.select_first()

    def _has_muc_service(self, account: str) -> bool:
        client = app.get_client(account)
        return client.get_module("MUC").service_jid is not None

    def _get_muc_service_jid(self) -> str:
        assert self._account is not None
        client = app.get_client(self._account)
        service_jid = client.get_module("MUC").service_jid
        if service_jid is None:
            return "muc.example.org"
        return str(service_jid)

    def _on_account_changed(
        self, dropdown: GajimDropDown[str], _param: GObject.ParamSpec
    ) -> None:
        self._account = dropdown.get_selected_key()
        log.debug("Set current account to: %s", self._account)
        if self._account is None:
            # Happens when we update the dropdown entries after an account
            # changes its state.
            return

        self._unset_error()
        self._unset_info()

        has_muc_service = self._has_muc_service(self._account)
        if has_muc_service:
            server = self._get_muc_service_jid()
            room_jid = f"{get_random_muc_localpart()}@{server}"
            self._address_entry_row.set_text(room_jid)
        else:
            self._address_entry_row.set_text("")

        if not has_muc_service:
            self._set_info(
                _(
                    "Your server does not offer a group chat service. "
                    "Please specify the address of a different server."
                )
            )

    def _is_jid_valid(self, text: str) -> bool:
        if not text:
            return True

        try:
            jid = validate_jid(text)
            if jid.resource:
                raise ValueError

        except ValueError:
            return False

        return True

    def _set_processing_state(self, enabled: bool) -> None:
        self._spinner.set_visible(enabled)
        self._account_group.set_sensitive(not enabled)
        self._settings_listbox.set_sensitive(not enabled)
        self._create_button.set_sensitive(not enabled)

    def _unset_info(self) -> None:
        self._info_label.set_visible(False)

    def _set_info(self, text: str) -> None:
        self._info_label.set_text(text)
        self._info_label.set_visible(True)

    def _unset_error(self) -> None:
        self._error_label.set_visible(False)
        self._create_button.set_sensitive(True)

    def _set_error(self, text: str) -> None:
        self._error_label.set_text(text)
        self._error_label.set_visible(True)
        self._create_button.set_sensitive(False)

    def _set_error_from_error(self, error: StanzaError) -> None:
        condition = error.condition or ""
        if condition == "gone":
            condition = "already-exists"
        text = MUC_DISCO_ERRORS.get(condition, to_user_string(error))
        self._set_error(text)

    def _set_error_from_error_code(self, error_code: str) -> None:
        self._set_error(MUC_DISCO_ERRORS[error_code])

    def _on_address_entry_changed(self, entry: Gtk.Entry) -> None:
        self._unset_error()

    def _on_create_clicked(self, _button: Gtk.Button) -> None:
        if self._account is None or not app.account_is_available(self._account):
            InformationAlertDialog(
                _("You are offline"),
                _("You have to be connected to create a group chat."),
            )
            return

        room_jid = self._address_entry_row.get_text()
        if not room_jid:
            room_jid = f"{get_random_muc_localpart()}@{self._get_muc_service_jid()}"

        if not self._is_jid_valid(room_jid):
            self._set_error(_("Invalid Address"))
            return

        self._set_processing_state(True)
        client = app.get_client(self._account)
        client.get_module("Discovery").disco_info(
            room_jid, callback=self._disco_info_received
        )

    @ensure_not_destroyed
    def _disco_info_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            if error.condition == "item-not-found":
                assert error.jid is not None
                self._create_muc(error.jid)
                return
            self._set_error_from_error(error)

        else:
            self._set_error_from_error_code(
                "already-exists" if result.is_muc else "not-muc-service"
            )

        self._set_processing_state(False)

    def _create_muc(self, room_jid: JID) -> None:
        name = self._name_entry_row.get_text()
        description = self._description_entry_row.get_text()
        is_public = self._public_radio.get_active()

        config = {
            # XEP-0045 options
            "muc#roomconfig_roomname": name,
            "muc#roomconfig_roomdesc": description,
            "muc#roomconfig_publicroom": is_public,
            "muc#roomconfig_membersonly": not is_public,
            "muc#roomconfig_whois": "moderators" if is_public else "anyone",
            "muc#roomconfig_changesubject": not is_public,
            # Ejabberd options
            "public_list": is_public,
        }

        # Create new group chat by joining
        assert self._account is not None

        if app.window.chat_exists(self._account, room_jid):
            log.error("Trying to create groupchat which is already added as chat")
            self.close()
            return

        client = app.get_client(self._account)
        client.get_module("MUC").create(room_jid, config)

        self.close()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging
import os
import sys
import zipfile
from collections.abc import Callable
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location
from pathlib import Path
from shutil import move
from shutil import rmtree
from types import TracebackType

from gajim.common import app
from gajim.common import configpaths
from gajim.common import modules
from gajim.common.client import Client
from gajim.common.exceptions import PluginsystemError
from gajim.common.i18n import _
from gajim.common.modules import BaseModule
from gajim.common.util.classes import Singleton

from .events import PluginAdded
from .events import PluginRemoved
from .gajimplugin import GajimPlugin
from .helpers import GajimPluginActivateException
from .helpers import is_shipped_plugin
from .manifest import PluginManifest

log = logging.getLogger("gajim.p.manager")

RmErrorT = tuple[type[BaseException], BaseException, TracebackType]

FIELDS = ("name", "short_name", "version", "description", "authors", "homepage")


class PluginManager(metaclass=Singleton):
    """
    Main plug-in management class.
    """

    def __init__(self):
        self.plugins: list[GajimPlugin] = []
        """
        Detected plugin classes.

        Each class object in list is `GajimPlugin` subclass.
        """
        self.active_plugins: list[GajimPlugin] = []
        """
        Instance objects of active plugins.

        These are object instances of classes held `plugins`, but only those
        that were activated.
        """
        self.gui_extension_points: dict[str, Any] = {}
        """
        Registered GUI extension points.
        """

        self.gui_extension_points_handlers: dict[str, Any] = {}
        """
        Registered handlers of GUI extension points.
        """

        self.encryption_plugins: dict[str, GajimPlugin] = {}
        """
        Registered names with instances of encryption Plugins.
        """

        self.update_plugins()
        self._load_manifests()

    def update_plugins(
        self,
        replace: bool = True,
        activate: bool = False,
        plugin_name: str | None = None,
    ) -> list[str]:
        """
        Move plugins from the downloaded folder to the user plugin folder

        :param replace: replace plugin files if they already exist.
        :param activate: load and activate the plugin
        :param plugin_name: if provided, update only this plugin
        :return: list of updated plugins (files have been installed)
        """
        updated_plugins: list[str] = []
        user_dir = configpaths.get("PLUGINS_USER")
        dl_dir = configpaths.get("PLUGINS_DOWNLOAD")
        to_update = [plugin_name] if plugin_name else next(os.walk(dl_dir))[1]
        for directory in to_update:
            src_dir = dl_dir / directory
            dst_dir = user_dir / directory
            try:
                if dst_dir.exists():
                    if not replace:
                        continue
                    self.delete_plugin_files(dst_dir)
                move(src_dir, dst_dir)
            except Exception:
                log.exception(
                    "Upgrade of plugin %s failed. "
                    'Impossible to move files from "%s" to "%s"',
                    directory,
                    src_dir,
                    dst_dir,
                )
                continue
            updated_plugins.append(directory)
            if activate:
                manifest = self._load_manifest(Path(dst_dir))
                if manifest is None or not manifest.is_usable:
                    log.warning("Error while updating plugin")
                    continue

                self._add_plugin(manifest, activate=True)
        return updated_plugins

    def init_plugins(self) -> None:
        for plugin in self.plugins:
            if not app.settings.get_plugin_setting(
                plugin.manifest.short_name, "active"
            ):
                continue
            if not plugin.activatable:
                continue

            try:
                self.activate_plugin(plugin, init=True)
            except GajimPluginActivateException as error:
                plugin.activatable = False
                plugin.available_text = str(error)

    def _load_plugin_module(self, manifest: PluginManifest) -> type[GajimPlugin] | None:
        assert manifest.path is not None
        module_path = manifest.path / "__init__.py"
        if not module_path.exists():
            # On Windows we only ship compiled files
            module_path = manifest.path / "__init__.pyc"

        module_name = manifest.path.stem

        try:
            spec = spec_from_file_location(module_name, module_path)
            if spec is None:
                return None
            module = module_from_spec(spec)
            sys.modules[spec.name] = module
            assert spec.loader is not None
            spec.loader.exec_module(module)
        except Exception as error:
            log.warning(
                "%s: Error while loading module: %s", manifest.short_name, error
            )
            return None

        for module_attr_name in dir(module):
            module_attr = getattr(module, module_attr_name)
            if issubclass(module_attr, GajimPlugin):
                module_attr.manifest = manifest
                module_attr.__path__ = str(manifest.path)
                return module_attr
        return None

    def _add_plugin(
        self, manifest: PluginManifest, activate: bool = False
    ) -> GajimPlugin | None:
        plugin_class = self._load_plugin_module(manifest)
        if plugin_class is None:
            return None

        if manifest in [p.manifest for p in self.plugins]:
            log.info(
                "Not loading plugin %s v %s. Plugin already loaded.",
                manifest.short_name,
                manifest.version,
            )
            return None

        try:
            plugin_obj = plugin_class()
        except Exception:
            log.exception("Error while loading a plugin")
            return None

        if manifest.short_name not in app.settings.get_plugins():
            # Enable plugins automatically when we first see them
            # if they are shipped by the distro
            app.settings.set_plugin_setting(
                manifest.short_name, "active", manifest.is_shipped
            )

        self.plugins.append(plugin_obj)
        plugin_obj.active = False

        if activate:
            try:
                self.activate_plugin(plugin_obj)
            except GajimPluginActivateException:
                pass

        app.ged.raise_event(PluginAdded(manifest=manifest))

        return plugin_obj

    def _remove_plugin(self, plugin: GajimPlugin) -> None:
        """
        removes the plugin from the plugin list and deletes all loaded modules
        from sys. This way we will have a fresh start when the plugin gets added
        again.
        """
        if plugin.active:
            self.deactivate_plugin(plugin)

        self.plugins.remove(plugin)

        # remove modules from cache
        base_package = plugin.__module__.split(".")[0]
        # get the subpackages/-modules of the base_package. Add a dot to the
        # name to avoid name problems (removing module_abc if base_package is
        # module_ab)
        modules_to_remove = [
            module for module in sys.modules if module.startswith(f"{base_package}.")
        ]
        # remove the base_package itself
        if base_package in sys.modules:
            modules_to_remove.append(base_package)

        for module_to_remove in modules_to_remove:
            del sys.modules[module_to_remove]

    def get_active_plugin(self, plugin_name: str) -> GajimPlugin | None:
        for plugin in self.active_plugins:
            if plugin.manifest.short_name == plugin_name:
                return plugin
        return None

    def get_plugin(self, short_name: str) -> GajimPlugin | None:
        for plugin in self.plugins:
            if plugin.manifest.short_name == short_name:
                return plugin
        return None

    def extension_point(self, gui_extpoint_name: str, *args: Any) -> None:
        """
        Invokes all handlers (from plugins) for a particular extension point,
        but doesn't add it to collection for further processing.
        For example if you pass a message for encryption via extension point to
        a plugin, its undesired that the call is stored and replayed on
        activating the plugin. For example after an update.

        :param gui_extpoint_name: name of GUI extension point.
        :param args: parameters to be passed to extension point handlers
                (typically and object that invokes `gui_extension_point`;
                however, this can be practically anything)
        :type args: tuple
        """

        self._execute_all_handlers_of_gui_extension_point(gui_extpoint_name, *args)

    def gui_extension_point(self, gui_extpoint_name: str, *args: Any) -> None:
        """
        Invokes all handlers (from plugins) for particular GUI extension point
        and adds it to collection for further processing (eg. by plugins not
        active yet).

        :param gui_extpoint_name: name of GUI extension point.
        :param args: parameters to be passed to extension point handlers
                (typically and object that invokes `gui_extension_point`;
                however, this can be practically anything)
        :type args: tuple
        """

        self._add_gui_extension_point_call_to_list(gui_extpoint_name, *args)
        self._execute_all_handlers_of_gui_extension_point(gui_extpoint_name, *args)

    def remove_gui_extension_point(self, gui_extpoint_name: str, *args: Any) -> None:
        """
        Removes GUI extension point from collection held by `PluginManager`.

        From this point this particular extension point won't be visible
        to plugins (eg. it won't invoke any handlers when plugin is activated).

        GUI extension point is removed completely (there is no way to recover it
        from inside `PluginManager`).

        Removal is needed when instance object that given extension point was
        connect with is destroyed (eg. ChatControl is closed or context menu
        is hidden).

        Each `PluginManager.gui_extension_point` call should have a call of
        `PluginManager.remove_gui_extension_point` related to it.

        :note: in current implementation different arguments mean different
                extension points. The same arguments and the same name mean
                the same extension point.

        :param gui_extpoint_name: name of GUI extension point.
        :param args: arguments that `PluginManager.gui_extension_point` was
                called with for this extension point. This is used (along with
                extension point name) to identify element to be removed.
        :type args: tuple
        """
        if gui_extpoint_name in self.gui_extension_points:
            extension_points = list(self.gui_extension_points[gui_extpoint_name])
            for ext_point in extension_points:
                if args[0] in ext_point:
                    self.gui_extension_points[gui_extpoint_name].remove(ext_point)

        if gui_extpoint_name not in self.gui_extension_points_handlers:
            return

        for handlers in self.gui_extension_points_handlers[gui_extpoint_name]:
            disconnect_handler = handlers[1]
            if disconnect_handler is not None:
                disconnect_handler(args[0])

    def _add_gui_extension_point_call_to_list(
        self, gui_extpoint_name: str, *args: Any
    ) -> None:
        """
        Adds GUI extension point call to list of calls.

        This is done only if such call hasn't been added already
        (same extension point name and same arguments).

        :note: This is assumption that GUI extension points are different only
        if they have different name or different arguments.

        :param gui_extpoint_name: GUI extension point name used to identify it
                by plugins.
        :param args: parameters to be passed to extension point handlers
                (typically and object that invokes `gui_extension_point`;
                however, this can be practically anything)
        :type args: tuple

        """
        if (
            gui_extpoint_name not in self.gui_extension_points
            or args not in self.gui_extension_points[gui_extpoint_name]
        ):
            self.gui_extension_points.setdefault(gui_extpoint_name, []).append(args)

    def _execute_all_handlers_of_gui_extension_point(
        self, gui_extpoint_name: str, *args: Any
    ) -> None:
        if gui_extpoint_name in self.gui_extension_points_handlers:
            for handlers in self.gui_extension_points_handlers[gui_extpoint_name]:
                try:
                    handlers[0](*args)
                except Exception:
                    log.warning("Error executing %s", handlers[0], exc_info=True)

    def _register_events_handlers_in_ged(self, plugin: GajimPlugin) -> None:
        for event_name, handler in plugin.events_handlers.items():
            priority = handler[0]
            handler_function = handler[1]
            app.ged.register_event_handler(event_name, priority, handler_function)

    def _remove_events_handler_from_ged(self, plugin: GajimPlugin) -> None:
        for event_name, handler in plugin.events_handlers.items():
            priority = handler[0]
            handler_function = handler[1]
            app.ged.remove_event_handler(event_name, priority, handler_function)

    def _remove_name_from_encryption_plugins(self, plugin: GajimPlugin) -> None:
        if plugin.encryption_name:
            self.encryption_plugins.pop(plugin.encryption_name, None)

    def _register_modules_with_handlers(self, plugin: GajimPlugin) -> None:
        if not hasattr(plugin, "modules"):
            return
        for client in app.get_clients():
            for module in plugin.modules:
                instance, name = module.get_instance(client)
                modules.register_single_module(client, instance, name)

                for handler in instance.handlers:
                    client.connection.register_handler(handler)

    def _unregister_modules_with_handlers(self, plugin: GajimPlugin) -> None:
        if not hasattr(plugin, "modules"):
            return
        for client in app.get_clients():
            for module in plugin.modules:
                try:
                    instance = cast(BaseModule, client.get_module(module.name))
                except Exception:
                    log.exception("Error while unregistering module")
                    continue

                modules.unregister_single_module(client, module.name)

                for handler in instance.handlers:
                    client.connection.unregister_handler(handler)

    def activate_plugin(self, plugin: GajimPlugin, init: bool = False) -> None:
        """
        :param plugin: plugin to be activated
        :type plugin: class object of `GajimPlugin` subclass
        """
        if plugin.active or not plugin.activatable:
            return

        try:
            self._add_gui_extension_points_handlers_from_plugin(plugin)
            self._add_encryption_name_from_plugin(plugin)
            self._handle_all_gui_extension_points_with_plugin(plugin)
            self._register_events_handlers_in_ged(plugin)
            self._register_modules_with_handlers(plugin)

            self.active_plugins.append(plugin)

            plugin.activate()
        except Exception as error:
            self.deactivate_plugin(plugin)
            log.exception("Error while activating plugin")
            raise GajimPluginActivateException(str(error))

        app.settings.set_plugin_setting(plugin.manifest.short_name, "active", True)
        plugin.active = True

        if not init:
            app.commands.init()

    def deactivate_plugin(self, plugin: GajimPlugin) -> None:
        self._remove_gui_extension_points_handlers_from_plugin(plugin)

        # detaching plug-in from handler GUI extension points (calling
        # cleaning up method that must be provided by plug-in developer
        # for each handled GUI extension point)
        for (
            gui_extpoint_name,
            gui_extpoint_handlers,
        ) in plugin.gui_extension_points.items():
            if gui_extpoint_name in self.gui_extension_points:
                for gui_extension_point_args in self.gui_extension_points[
                    gui_extpoint_name
                ]:
                    handler = gui_extpoint_handlers[1]
                    if handler:
                        try:
                            handler(*gui_extension_point_args)
                        except Exception:
                            log.warning("Error executing %s", handler, exc_info=True)

        self._remove_events_handler_from_ged(plugin)
        self._remove_name_from_encryption_plugins(plugin)

        # deactivate() must be before _unregister_modules_with_handlers(),
        # because plugin.deactivate() may want to use the module
        plugin.deactivate()

        self._unregister_modules_with_handlers(plugin)
        try:
            self.active_plugins.remove(plugin)
        except ValueError:
            pass

        app.settings.set_plugin_setting(plugin.manifest.short_name, "active", False)
        plugin.active = False
        app.commands.init()

    def _add_gui_extension_points_handlers_from_plugin(
        self, plugin: GajimPlugin
    ) -> None:
        for (
            gui_extpoint_name,
            gui_extpoint_handlers,
        ) in plugin.gui_extension_points.items():
            self.gui_extension_points_handlers.setdefault(gui_extpoint_name, []).append(
                gui_extpoint_handlers
            )

    def _remove_gui_extension_points_handlers_from_plugin(
        self, plugin: GajimPlugin
    ) -> None:
        for (
            gui_extpoint_name,
            gui_extpoint_handlers,
        ) in plugin.gui_extension_points.items():
            try:
                self.gui_extension_points_handlers[gui_extpoint_name].remove(
                    gui_extpoint_handlers
                )
            except ValueError:
                pass

    def _add_encryption_name_from_plugin(self, plugin: GajimPlugin) -> None:
        if plugin.encryption_name:
            self.encryption_plugins[plugin.encryption_name] = plugin

    def _handle_all_gui_extension_points_with_plugin(self, plugin: GajimPlugin) -> None:
        for (
            gui_extpoint_name,
            gui_extpoint_handlers,
        ) in plugin.gui_extension_points.items():
            if gui_extpoint_name in self.gui_extension_points:
                for gui_extension_point_args in self.gui_extension_points[
                    gui_extpoint_name
                ]:
                    handler = gui_extpoint_handlers[0]
                    if handler:
                        try:
                            handler(*gui_extension_point_args)
                        except Exception:
                            log.warning("Error executing %s", handler, exc_info=True)

    def register_modules_for_account(self, client: Client) -> None:
        """
        A new account has been added, register modules
        of all active plugins
        """
        for plugin in self.plugins:
            if not plugin.active:
                continue

            if not hasattr(plugin, "modules"):
                continue

            for module in plugin.modules:
                instance, name = module.get_instance(client)
                modules.register_single_module(client, instance, name)

                for handler in instance.handlers:
                    client.connection.register_handler(handler)

    @staticmethod
    def _load_manifest(plugin_path: Path) -> PluginManifest | None:
        try:
            return PluginManifest.from_path(plugin_path)
        except Exception as error:
            log.debug("Unable to load manifest: %s", error)
            return None

    def _load_manifests(self) -> None:
        manifests: dict[str, PluginManifest] = {}
        for plugin_dir in configpaths.get_plugin_dirs():
            if not plugin_dir.is_dir():
                continue

            for plugin_path in plugin_dir.iterdir():
                manifest = self._load_manifest(plugin_path)
                if manifest is None:
                    continue

                if not manifest.is_usable:
                    log.info(
                        "Skipped plugin with not suitable requirements: %s %s",
                        manifest.short_name,
                        manifest.requirements,
                    )
                    continue

                same_plugin = manifests.get(manifest.short_name)
                if same_plugin is not None:
                    if same_plugin.version > manifest.version:
                        continue

                log.info("Found plugin %s %s", manifest.short_name, manifest.version)
                manifests[manifest.short_name] = manifest

        for manifest in manifests.values():
            self._add_plugin(manifest)

    def install_from_zip(
        self, zip_filename: str, overwrite: bool = False
    ) -> GajimPlugin | None:
        """
        Install plugin from zip and return plugin
        """
        try:
            zip_file = zipfile.ZipFile(zip_filename)
        except zipfile.BadZipfile:
            # it is not zip file
            raise PluginsystemError(_("Archive corrupted"))
        except OSError:
            raise PluginsystemError(_("Archive empty"))

        if zip_file.testzip():
            # CRC error
            raise PluginsystemError(_("Archive corrupted"))

        dirs: list[str] = []
        manifest = None
        for filename in zip_file.namelist():
            if filename.startswith((".", "/")) or ("/" not in filename):
                # members not safe
                raise PluginsystemError(_("Archive is malformed"))
            if filename.endswith("/") and filename.find("/", 0, -1) < 0:
                dirs.append(filename.strip("/"))
            if "plugin-manifest.json" in filename.split("/")[1]:
                manifest = True
        if not manifest:
            return None
        if len(dirs) > 1:
            raise PluginsystemError(_("Archive is malformed"))

        plugin_name = dirs[0]
        user_dir = configpaths.get("PLUGINS_USER")
        plugin_path = user_dir / plugin_name

        if plugin_path.exists():
            # Plugin dir already exists
            if not overwrite:
                raise PluginsystemError(_("Plugin already exists"))
            plugin = self.get_plugin_by_path(str(plugin_path))
            assert isinstance(plugin, GajimPlugin)
            self.uninstall_plugin(plugin)

        zip_file.extractall(user_dir)
        zip_file.close()

        manifest = self._load_manifest(plugin_path)
        if manifest is None or not manifest.is_usable:
            log.warning("Error while installing from zip")
            rmtree(plugin_path)
            raise PluginsystemError(_("Installation failed"))

        return self._add_plugin(manifest)

    def delete_plugin_files(self, plugin_path: Path) -> None:
        def _on_error(func: Callable[..., Any], path: Path, error: RmErrorT) -> None:
            if func is os.path.islink:
                # if symlink
                os.unlink(path)
                return
            # access is denied or other
            raise PluginsystemError(str(error[1]))

        rmtree(plugin_path, False, _on_error)  # type: ignore

    def uninstall_plugin(self, plugin: GajimPlugin) -> None:
        """
        Deactivate and remove plugin from `plugins` list
        """
        self._remove_plugin(plugin)
        self.delete_plugin_files(Path(plugin.__path__))
        if not is_shipped_plugin(Path(plugin.__path__)):
            path = configpaths.get("PLUGINS_BASE") / plugin.manifest.short_name
            if path.exists():
                self.delete_plugin_files(path)

        path = configpaths.get("PLUGINS_DOWNLOAD") / plugin.manifest.short_name
        if path.exists():
            self.delete_plugin_files(path)

        app.settings.remove_plugin(plugin.manifest.short_name)

        app.ged.raise_event(PluginRemoved(manifest=plugin.manifest))

    def get_plugin_by_path(self, plugin_dir: str) -> GajimPlugin | None:
        for plugin in self.plugins:
            if plugin.__path__ in plugin_dir:
                return plugin
        return None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Message Util module

from __future__ import annotations

from typing import Literal

import base64
from datetime import datetime
from datetime import UTC

import nbxmpp.structs
from nbxmpp.modules.security_labels import SecurityLabel
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import MessageProperties

from gajim.common import app
from gajim.common.const import EME_MESSAGES
from gajim.common.const import EME_PROTOCOLS
from gajim.common.const import IMAGE_MIME_TYPES
from gajim.common.const import VALUE_MISSING
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.structs import MUCData
from gajim.common.structs import ReplyData
from gajim.common.util.uri import DataUri

UNKNOWN_MESSAGE = _("Message content unknown")


def get_eme_message(eme_data: nbxmpp.structs.EMEData) -> str:
    try:
        return EME_MESSAGES[eme_data.namespace]
    except KeyError:
        encryption_name = EME_PROTOCOLS.get(eme_data.namespace)
        if not encryption_name:
            encryption_name = eme_data.name or eme_data.namespace
        return EME_MESSAGES["fallback"] % encryption_name


def get_eme_protocol(eme_data: nbxmpp.structs.EMEData) -> str:
    return EME_PROTOCOLS.get(eme_data.namespace, "Unknown")


def get_chat_type_and_direction(
    muc_data: MUCData | None, own_jid: JID, properties: MessageProperties
) -> tuple[MessageType, ChatDirection]:

    assert properties.jid is not None
    if properties.type.is_groupchat:
        assert muc_data is not None
        direction = ChatDirection.INCOMING
        if muc_data.occupant_id is not None:
            if muc_data.occupant_id == properties.occupant_id:
                direction = ChatDirection.OUTGOING

        elif muc_data.nick == properties.jid.resource:
            direction = ChatDirection.OUTGOING

        return MessageType.GROUPCHAT, direction

    assert properties.from_ is not None
    if properties.from_.bare_match(own_jid):
        direction = ChatDirection.OUTGOING
    else:
        direction = ChatDirection.INCOMING

    if properties.is_muc_pm:
        return MessageType.PM, direction

    return MessageType.CHAT, direction


def get_real_jid(
    properties: MessageProperties,
    contact: GroupchatParticipant,
) -> JID | None:

    if properties.is_mam_message:
        if properties.muc_user is None:
            return None
        return properties.muc_user.jid

    real_jid = contact.real_jid
    if real_jid is None:
        return None
    return real_jid.new_as_bare()


def get_occupant_info(
    account: str,
    remote_jid: JID,
    own_bare_jid: JID,
    direction: ChatDirection,
    m_type: MessageType,
    timestamp: datetime,
    contact: GroupchatParticipant,
    properties: MessageProperties,
) -> mod.Occupant | None:

    assert properties.jid is not None
    if properties.jid.is_bare:
        return None

    if direction == ChatDirection.OUTGOING:
        real_jid = own_bare_jid
    else:
        real_jid = get_real_jid(properties, contact)

    occupant_id = get_occupant_id(contact, properties) or real_jid
    if occupant_id is None:
        return None

    if m_type == MessageType.PM and direction == ChatDirection.OUTGOING:
        # Outgoing PMs which we receive as carbons or via MAM
        # have our bound jid in the from attribute, not the joined MUC
        # resource. This means for these kind of messages the nickname
        # cannot be determined
        resource = None
    else:
        resource = properties.jid.resource
        assert resource is not None

    occupant = mod.Occupant(
        account_=account,
        remote_jid_=remote_jid.new_as_bare(),
        id=str(occupant_id),
        real_remote_jid_=real_jid or VALUE_MISSING,
        nickname=resource,
        updated_at=timestamp,
    )

    if contact.avatar_sha is not None:
        # avatar_sha is only available if we have presence from this contact
        # We don’t want to overwrite the previous avatar
        occupant.avatar_sha = contact.avatar_sha

    return occupant


def get_occupant_id(
    contact: GroupchatParticipant,
    properties: MessageProperties,
) -> str | None:

    if not properties.occupant_id:
        return None

    if contact.room.supports(Namespace.OCCUPANT_ID):
        return properties.occupant_id
    return None


def get_message_timestamp(properties: MessageProperties) -> datetime:
    timestamp = properties.timestamp
    if properties.mam is not None:
        timestamp = properties.mam.timestamp
    return datetime.fromtimestamp(timestamp, tz=UTC)


def convert_message_type(type_: MessageType) -> Literal["chat", "groupchat"]:
    if type_ in (MessageType.CHAT, MessageType.PM):
        return "chat"
    return "groupchat"


def get_nickname_from_message(account: str, message: mod.Message) -> tuple[str, str]:
    # This method is tailored for the ChatListRow
    def _get_nickname() -> str:
        if occupant := message.occupant:
            if nickname := occupant.nickname:
                return nickname

        if message.resource is not None:
            return message.resource

        nickname = message.remote.jid.localpart
        assert nickname is not None
        return nickname

    match message.type:
        case MessageType.CHAT:
            if message.direction == ChatDirection.INCOMING:
                return "", ""

            nickname = app.get_client(account).get_own_contact().name
            return _("Me"), nickname

        case MessageType.PM:
            if message.direction == ChatDirection.INCOMING:
                return "", ""
            return _("Me"), _get_nickname()

        case MessageType.GROUPCHAT:
            if message.direction == ChatDirection.INCOMING:
                return _get_nickname(), ""
            return _("Me"), _get_nickname()

        case _:
            raise ValueError(f"Unknown type: {message.type}")


def get_security_label(
    account: str, remote_jid: JID, timestamp: datetime, label: SecurityLabel | None
) -> mod.SecurityLabel | None:
    if label is None:
        return None

    displaymarking = label.displaymarking
    if displaymarking is None:
        return None

    return mod.SecurityLabel(
        account_=account,
        remote_jid_=remote_jid,
        label_hash=label.get_label_hash(),
        displaymarking=displaymarking.name,
        fgcolor=displaymarking.fgcolor,
        bgcolor=displaymarking.bgcolor,
        updated_at=timestamp,
    )


def get_reply(data: nbxmpp.structs.ReplyData | ReplyData | None) -> mod.Reply | None:
    if data is None:
        return None

    return mod.Reply(id=data.id, to=data.to)


def get_open_graph(
    open_graph: dict[str, nbxmpp.structs.OpenGraphData] | dict[str, OpenGraphData],
) -> list[mod.OpenGraph]:

    og_data: list[mod.OpenGraph] = []
    for url, data in open_graph.items():
        if not data.title:
            continue

        if isinstance(data, OpenGraphData):
            image_type, image_bytes = None, None
            if thumbnail := data.thumbnail:
                image_type, image_bytes = thumbnail.type, thumbnail.data
        else:
            image_type, image_bytes = get_thumbnail(data.image)

        og_data.append(
            mod.OpenGraph(
                about=url,
                title=data.title,
                description=data.description or None,
                image_type=image_type,
                image_bytes=image_bytes,
            )
        )

    return og_data


def get_thumbnail(image_uri: str | None) -> tuple[str, bytes] | tuple[None, None]:
    if not image_uri:
        return None, None

    def _parse() -> tuple[str, bytes]:
        datauri = DataUri.from_string(image_uri)
        if not datauri.is_base64:
            raise ValueError

        image_bytes = base64.b64decode(datauri.data)
        if not image_bytes:
            raise ValueError

        if datauri.media_type not in IMAGE_MIME_TYPES:
            raise ValueError

        if len(image_bytes) > 50_000:
            raise ValueError

        return datauri.media_type, image_bytes

    try:
        return _parse()
    except Exception:
        return None, None


def successfully_decrypted(properties: MessageProperties) -> bool:
    return properties.eme is None or properties.encrypted is not None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GtkSource

from gajim.common.const import Direction

from gajim.gtk.completion.base import BaseCompletionListItem
from gajim.gtk.completion.base import BaseCompletionProvider
from gajim.gtk.completion.base import BaseCompletionViewItem

log = logging.getLogger("gajim.gtk.chat_action_processor")


class CompletionPopover(Gtk.Popover):
    __gsignals__ = {
        "completion-picked": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (str,),
        )
    }

    def __init__(self, message_input: GtkSource.View) -> None:
        Gtk.Popover.__init__(
            self,
            autohide=False,
            has_arrow=False,
            position=Gtk.PositionType.TOP,
        )
        self.add_css_class("completion-popover")
        self.set_offset(100, 0)

        self._message_input = message_input
        self._provider: BaseCompletionProvider | None = None
        self._widget_cls: type[BaseCompletionViewItem[Any]] | None = None

        controller = Gtk.EventControllerKey(
            propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        controller.connect("key-pressed", self._on_key_pressed)
        self._message_input.add_controller(controller)

        box = Gtk.Box(spacing=6, orientation=Gtk.Orientation.VERTICAL)

        self._view = CompletionListView(
            model=Gtk.SingleSelection(), single_click_activate=True, can_focus=False
        )
        self._view.connect("activate", self._on_item_activated)
        self._view.connect("extended-activate", self._on_extended_item_activated)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup)
        factory.connect("bind", self._on_factory_bind)
        factory.connect("unbind", self._on_factory_unbind)
        self._view.set_factory(factory)

        self._header_label = Gtk.Label(halign=Gtk.Align.START, xalign=0)
        self._header_label.add_css_class("dimmed")
        box.append(self._header_label)
        box.append(self._view)

        self.set_child(box)

    def set_provider(self, provider: BaseCompletionProvider) -> None:
        if self._provider is provider:
            return

        self._provider = provider
        model, widget_cls = self._provider.get_model()
        self._header_label.set_label(self._provider.name)
        self._widget_cls = widget_cls
        for name in self._view.get_css_classes():
            if name.endswith("-completion"):
                self._view.remove_css_class(name)
        self._view.add_css_class(widget_cls.css_class)
        selection_model = cast(Gtk.SingleSelection, self._view.get_model())
        assert selection_model is not None
        selection_model.set_model(model)

    def _on_factory_setup(
        self,
        _factory: Gtk.SignalListItemFactory,
        list_item: Gtk.ListItem,
    ) -> None:
        assert self._widget_cls is not None
        list_item.set_child(self._widget_cls())  # type: ignore

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = cast(BaseCompletionViewItem[Any], list_item.get_child())
        obj = list_item.get_item()
        view_item.bind(obj)

    @staticmethod
    def _on_factory_unbind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = cast(BaseCompletionViewItem[Any], list_item.get_child())
        view_item.unbind()

    def _on_item_activated(self, list_view: Gtk.ListView, position: int) -> None:
        self.popdown()
        model = cast(Gtk.SingleSelection, list_view.get_model())
        assert model is not None
        item = cast(BaseCompletionListItem, model.get_item(position))
        self.emit("completion-picked", item.get_text())

    def _on_extended_item_activated(
        self, list_view: Gtk.ListView, complete_string: str
    ) -> None:
        self.popdown()
        self.emit("completion-picked", complete_string)

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if not self.get_visible():
            return Gdk.EVENT_PROPAGATE

        if keyval == Gdk.KEY_Up:
            self._select(Direction.PREV)
            return Gdk.EVENT_STOP

        if keyval == Gdk.KEY_Down:
            self._select(Direction.NEXT)
            return Gdk.EVENT_STOP

        if keyval in (Gdk.KEY_Left, Gdk.KEY_Right):
            self.popdown()
            return Gdk.EVENT_PROPAGATE

        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter, Gdk.KEY_ISO_Enter, Gdk.KEY_Tab):
            model = cast(Gtk.SingleSelection, self._view.get_model())
            assert model is not None
            position = model.get_selected()
            if position != Gtk.INVALID_LIST_POSITION:
                self._view.emit("activate", position)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _select(self, direction: Direction) -> None:
        model = cast(Gtk.SingleSelection, self._view.get_model())
        assert model is not None
        selected_pos = model.get_selected()
        if selected_pos == Gtk.INVALID_LIST_POSITION:
            return

        if direction == Direction.NEXT:
            new_pos = selected_pos + 1
        else:
            new_pos = selected_pos - 1

        if not 0 <= new_pos < model.get_n_items():
            return

        model.set_selected(new_pos)


class CompletionListView(Gtk.ListView):
    __gsignals__ = {
        "extended-activate": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (str,),
        )
    }



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import sys
from pathlib import Path

from gi.repository import GObject
from gi.repository.Gdk import Paintable

from gajim.common import app

try:
    from gi.repository import Gst
except Exception:
    if typing.TYPE_CHECKING:
        from gi.repository import Gst

import logging

from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.preview_animated_image_backend")


class AnimatedImageBackend(GObject.Object, SignalManager):
    __gtype_name__ = "AnimatedImageBackend"
    __gsignals__ = {
        "pipeline-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
        "playback-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(self, orig_path: Path, max_loops: int = 3) -> None:
        super().__init__()
        SignalManager.__init__(self)

        self._orig_path = orig_path

        self._pipeline_is_setup = False
        self._pipeline_setup_failed = False

        self._paintable: Paintable | None = None
        self._pipeline: Gst.Pipeline | None = None
        self._src: Gst.Element | None = None
        self._bus: Gst.Bus | None = None

        # Workaround a regression of Gst or Gtk on Windows,
        # wglShareLists() called by GstGL would fail due to ERROR_BUSY (0xAA)
        self._use_gl = sys.platform != "win32"

        self._loop_counter = 0
        self._max_loop_counts = max_loops

    @property
    def pipeline_is_setup(self) -> bool:
        return self._pipeline_is_setup

    @property
    def pipeline_setup_failed(self) -> bool:
        return self._pipeline_setup_failed

    @property
    def paintable(self) -> Paintable | None:
        return self._paintable

    def cleanup(self) -> None:
        self._cleanup()
        self._disconnect_all()
        app.check_finalize(self)

    def is_playing(self) -> bool:
        if self._pipeline is None:
            return False
        _, state, _ = self._pipeline.get_state(timeout=100)
        return state == Gst.State.PLAYING

    def play(self) -> None:
        assert self._pipeline is not None
        self._pipeline.set_state(Gst.State.PLAYING)

    def pause(self) -> None:
        assert self._pipeline is not None
        self._loop_counter = 0
        self._pipeline.set_state(Gst.State.PAUSED)
        self._pipeline.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, 0)
        self.emit("playback-changed", False)

    def setup_pipeline(self) -> None:
        try:
            self._setup_pipeline()
        except Exception:
            log.exception("Unable to play animated image")
            self._pipeline_setup_failed = True
            self.emit("pipeline-changed", False)
        else:
            self._pipeline_is_setup = True
            self.emit("pipeline-changed", True)

    def _setup_pipeline(self) -> None:
        if self._pipeline_is_setup:
            return

        pipeline_parts = {
            "src": "filesrc name=src",
            "decode": "decodebin name=decodebin",
            "gl": "glupload name=glupload ! glcolorconvert" if self._use_gl else "",
            "queue": "queue name=queue",
            "sink": "gtk4paintablesink name=sink",
        }
        pipeline = " ! ".join(filter(None, pipeline_parts.values()))
        try:
            self._pipeline = typing.cast(Gst.Pipeline, Gst.parse_launch(pipeline))
        except Exception as e:
            self._pipeline_setup_failed = True
            self.emit("pipeline-changed", False)
            log.warning("Failed to setup pipeline: %s", e)
            return

        decodebin = typing.cast(Gst.Element, self._pipeline.get_by_name("decodebin"))
        self._connect(decodebin, "pad-added", self._on_pad_added)

        self._src = self._pipeline.get_by_name("src")
        assert self._src is not None
        self._src.set_property("location", str(self._orig_path))

        self._bus = self._pipeline.get_bus()
        self._bus.add_signal_watch()
        self._connect(self._bus, "message::eos", self._on_eos)
        self._connect(self._bus, "message::state-changed", self._on_state_changed)

        sink = self._pipeline.get_by_name("sink")
        assert sink is not None
        self._paintable = sink.get_property("paintable")

    def _on_pad_added(self, _bin: Gst.Bin, pad: Gst.Pad) -> None:
        assert pad is not None
        caps = pad.get_current_caps()
        if caps is None:
            return

        log.debug("Link decodebin with glupload")
        assert self._pipeline is not None
        if self._use_gl:
            elem = self._pipeline.get_by_name("glupload")
        else:
            elem = self._pipeline.get_by_name("queue")
        assert elem is not None
        sink_pad = elem.get_static_pad("sink")
        assert sink_pad is not None
        if not sink_pad.is_linked():
            pad.link(sink_pad)

    def _on_eos(self, _bus: Gst.Bus, _msg: Gst.Message) -> None:
        assert self._pipeline is not None
        self._pipeline.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, 0)
        self._loop_counter = (self._loop_counter + 1) % self._max_loop_counts
        if self._loop_counter == 0:
            self.pause()

    def _on_state_changed(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        assert self._pipeline is not None
        assert message is not None
        if message.src == self._pipeline:
            _, new, _ = message.parse_state_changed()
            if new == Gst.State.PLAYING:
                self.emit("playback-changed", True)

    def _cleanup(self) -> None:
        if self._pipeline is not None:
            self._pipeline.set_state(Gst.State.NULL)

        if self._pipeline is not None:
            self._pipeline.run_dispose()
            self._pipeline = None

        if self._bus is not None:
            self._bus.remove_signal_watch()
            self._bus = None

        self._pipeline_setup = False
        self._paintable = None
